/**
 *  Copyright (c) Members of the EGEE Collaboration. 2004.
 *  See http://www.eu-egee.org/partners/ for details on the copyright
 *  holders.
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 *  GLite Data Config Service : string conversion utilities
 *
 *  Authors: Paolo Tedesco <paolo.tedesco@cern.ch>
 *  Version info: $Id: StringConverter.hxx,v 1.5 2009/10/29 10:22:44 szamsu Exp $
 *  Release: $Name:  $
 *
 */

#ifndef GLITE_DATA_TRANSFER_LOAD_GENERATOR_STRINGCONVERTER_H_
#define GLITE_DATA_TRANSFER_LOAD_GENERATOR_STRINGCONVERTER_H_

// #include <boost/lexical_cast.hpp>
#include <sstream>
#include <map>
#include <algorithm>

namespace glite {
namespace config {

/**
 * StringConverter
 * convert a value into its string representation and viceversa.
 */
template <typename T>
class StringConverter{
public:
    
    /**
     * from_string
     * @param value [IN] string representation
     * @return value
     */
    static T from_string(const std::string& value){
        // NOTE: boost::lexical_cast gives problem in case there are leading 
        // spaces (which happens with xml text nodes), so not using it here.
        // return boost::lexical_cast<T>(value);
        std::stringstream ss;
        ss << value;
        T ret;
        ss >> ret;
        return ret;
    }
    
    /**
     * to_string
     * @param value [IN] value
     * @return string
     */
    static std::string to_string(T value){
        std::stringstream ss;
        ss << value;
        return ss.str();
    }
    
};

/**
 * StringConverter
 * specialization for bool parameters
 */
template <>
class StringConverter<bool>{
public:
    
    /**
     * from_string
     */
    static bool from_string(const std::string& value){
        std::string to_lower(value);
        std::transform(to_lower.begin(),to_lower.end(),to_lower.begin(),(int(*)(int))std::tolower);
        if (to_lower == "true" || to_lower == "yes") {
            return true;
        } else if (to_lower == "false" || to_lower == "no") {
            return false;
        } else {
            return false;
        }
    }

    /**
     * to_string
     */
    static std::string to_string(bool value){
        return value ? "true" : "false";
    }
        
};

/**
 * StringConverter
 * specialization for string parameters
 */
template <>
class StringConverter<std::string>
{
public:
    static std::string from_string(const std::string& value)
    {
        return value;
    }

    static std::string to_string(const std::string& value)
    {
        return value;
    }
};

/**
 * EnumConverter
 * base class for StringConverter specialization for enumerated types. See example.

struct State{
    enum Value{
        zero,
        one
    };
};

template<>
struct StringConverter<State::Value> : public EnumConverter<State::Value>{
    StringConverter() : EnumConverter<State::Value>(State::zero,"zero"){
        add(State::zero,"zero");
        add(State::one, "one");
    }
};
 */
template <typename Value> 
class EnumConverter{
public:
    
    /**
     * conversion from string
     */
    static Value from_string(const std::string& strval){
        return instance().string_to_value(strval);
    }
    
    /**
     * conversion to string
     */
    static std::string to_string(const Value value){
        return instance().value_to_string(value);
    }
    
    virtual ~EnumConverter(){}
    
protected:
    
    /**
     * instance
     * @return converter instance
     */
    static StringConverter<Value>& instance(){
        static StringConverter<Value> s_instance;
        return s_instance;
    }

    /**
     * string_to_value
     */
    Value string_to_value(const std::string& strval){
        typename StringToEnum::const_iterator it = m_StringToEnum.find(strval);
        if (m_StringToEnum.end() != it){
            return it->second;
        }
        return get_default_value();
    }
    
    /**
     * value_to_string
     */
    std::string value_to_string(const Value value){
        typename EnumToString::const_iterator it = m_EnumToString.find(value);
        if (m_EnumToString.end() != it){
            return it->second;
        }
        return get_default_string();
    }
    /**
     * constructor
     */
    EnumConverter() :
        _default_value((Value)-1),
        _default_string(""){
    }
    /**
     * constructor
     * @param default_value [IN]
     * @param default_string [IN]
     */
    EnumConverter(
        Value       default_value,
        std::string default_string) :
        _default_value(default_value),
        _default_string(default_string){
    }
    
    /**
     * add a value-string couple to internal maps
     */
    void add (const Value& v,const std::string& s){
        m_EnumToString[v] = s;
        m_StringToEnum[s] = v;
    }
    
    /**
     * get_default_string
     * can be overridden by derived classes e.g. to throw an exception
     * @return default string
     */
    virtual std::string get_default_string(){
        return _default_string;
    }
    /**
     * get_default_value
     * can be overridden by derived classes e.g. to throw an exception
     * @return default value    
     */
    virtual Value get_default_value(){
        return _default_value;
    }
    
    // maps
    typedef std::map<std::string,Value> StringToEnum;
    typedef std::map<Value,std::string> EnumToString;
    EnumToString m_EnumToString;
    StringToEnum m_StringToEnum;
    // default values    
    Value       _default_value;
    std::string _default_string;
    
};
    
} // namespace config
} // namespace glite

#endif

