/**
 * Copyright (c) Members of the EGEE Collaboration. 2004.
 * See http://www.eu-egee.org/partners/ for details on the copyright holders.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *     http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 *  Common Functions for the GLite Data Conf Service  module
 *
 *  Authors: Paolo Badino <paolo.badino@cern.ch>, Paolo Tedesco <paolo.tedesco@cern.ch>
 *  Version info: $Id: Functions.cxx,v 1.14 2009/10/29 10:22:44 szamsu Exp $
 *  Release: $Name:  $
 *
 */
#include "Functions.hxx" 
#include "Locations.hxx" 
#include "Echo.hxx"
#include <cstring>
#include <fstream>
#include <iostream>
#include <dirent.h>
#include <unistd.h>
#include <sys/file.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <sys/ipc.h>
#include <sys/sem.h>
#include <stdarg.h>
#include <errno.h>
#include <string.h>


namespace {
const char * PROC_DIR         = "/proc";
}
    
namespace glite {
namespace config{
    
/**
 * pid_of_running_proc
 * Get pid of process looking in the /proc directory
 */
int pid_of_running_proc(const Locations& locations, int name_len){
    
    const char * name = locations.serviceName().c_str();
    
    int pid = -1;
    
    // Search the process in the /proc dir
    DIR * dir = opendir(PROC_DIR);
    if(0 != dir){
        // Scan the directory
        struct dirent * dir_entry = 0;
        do{
            dir_entry = readdir(dir);
            if(0 != dir_entry){
                std::string proc_filename = (std::string)PROC_DIR + "/" + dir_entry->d_name + "/cmdline";
                // echo_debug("reading file '%s'",proc_filename.c_str());
                std::ifstream proc_file(proc_filename.c_str());
                if(proc_file.is_open()){
                    // Get the line
                    std::string line;
                    proc_file >> line;                  

                    // Check that the line start with the given name
                    if(0 == strncmp(name, line.c_str(),name_len-1)){
                        pid = atoi(dir_entry->d_name);
                        echo_debug("got running process pid from file '%s'",proc_filename.c_str()); 
			proc_file.sync();			
                        proc_file.close();
                        break;
                    }
		    proc_file.sync();
                    proc_file.close();
		    
                }
            }
        } while (dir_entry != 0);
        closedir(dir);
    } else {
        echo_debug("could not open directory '%s'",PROC_DIR);
    }
    if (-1 == pid){
        echo_debug("running process pid not found");
    }    
    return pid;
}

/**
 * pid_of_proc
 * Get pid of process, first looking at the .pid file and then in the /proc directory
 */
int pid_of_proc(const Locations& locations, int name_len){
    
    int pid = -1;
    const std::string& pid_filename = locations.pidFile();
    echo_debug("got pid filename '%s'",pid_filename.c_str());
    std::ifstream pid_file(pid_filename.c_str());
    if(pid_file.is_open()){
        echo_debug("reading pid from .pid file");
        pid_file >> pid;
        pid_file.sync();	
        pid_file.close();
        echo_debug("PID is: %d", pid);
    } else {
        echo_debug("reading pid of running process");
        pid = pid_of_running_proc(locations,name_len);
    }
    return pid;
}

/**
 * status_of_proc
 * Returns 0 is the process is running
 */
int status_of_proc(const Locations& locations, int name_len) {
  
    int status = -1;
    int pid = pid_of_proc(locations,name_len);
    echo_debug("process pid = %d",pid);
    if(-1 != pid){
        int check_pid_result = check_pid(pid);
        echo_debug("check pid returned %d",check_pid_result);
        if(0 == check_pid_result){
            status = 0;
        }
    }
    
    return status;
}

/**
 * check_pid
 * Check if pid is running
 */
int check_pid(int pid) {
    int result = -1;
    
    char pidstr[256];
    sprintf(pidstr,"%d",pid);
    std::string dir_name = (std::string)PROC_DIR + "/"+ pidstr;
    echo_debug("check_pid: dir_name = %s",dir_name.c_str());
    // Search the process in the /proc dir
    DIR * dir = opendir(dir_name.c_str());
    if(0 != dir){
        closedir(dir);
        // Try to open the cmdline file
        std::string fname = (std::string)PROC_DIR + "/"+ pidstr + "/cmdline";
        echo_debug("check_pid: fname = %s",fname.c_str());
        std::ifstream cmdline_file(fname.c_str());
        if(cmdline_file.is_open()){
            //try to read a char
            char c;
            cmdline_file.read(&c,1);
            if(cmdline_file.good()){
                result = 0;
            } else {
                result = -4;				
            }
        } else {
            result = -3;
        }
    } else {
        result = -2;
    }   
   return result;
}

/**
 * write_pidfile
 * Write the pid file for the given process
 * @param name [IN] the name of the process
 * @param name_len [IN] the meaningful length of the process name
 * @param location [IN] the location path used to create the pid file name
 * @param location_var [IN] the var location path used to create the pid file name
 * @param pid [IN] Optional, the id of the process. If 0 is specified, the pid 
 * is computed using pid_of_running_process
 * @return 0 if success
 */
int write_pidfile(const Locations& locations, int name_len, int pid){
    int result = -1;
    if(0 == pid){
        pid = pid_of_running_proc(locations,name_len);
    }
    if(pid > 0) {
        const std::string& pid_filename = locations.pidFile();
        std::ofstream pidfile(pid_filename.c_str(), std::ios::out | std::ios::trunc);
        if(pidfile.is_open()){
            pidfile << pid << std::endl;
            pidfile.close();
	    pidfile.flush();
            result = 0;
        }
    }
    return result;
}

/**
 * unlink_pidfile
 * Delete the pid file for the given service
 */
int unlink_pidfile(const Locations& locations, int name_len) {
    const std::string& pid_filename = locations.pidFile();
    int result = unlink(pid_filename.c_str());
    return result;
}

/**
 * proc_acquire_lock
 * Acquire a lock for the process. This method will fail if another instance is running 
 */
int proc_acquire_lock(const Locations& locations){
    const std::string& lock_filename = locations.lockFile();
    int filelock = open(lock_filename.c_str(), O_WRONLY | O_CREAT | O_APPEND, 0600);
    if (-1 != filelock){
        if(0 == flock(filelock,LOCK_EX | LOCK_NB )){
            return filelock;
        } else {
            close(filelock);
        }
    }
    return -1;
}

/**
 * proc_release_lock
 * Release a lock for the process
 */
int proc_release_lock(
    const Locations&  locations, 
    int         name_len, 
    int         filelock){
    if(-1 != filelock){
        flock(filelock,LOCK_UN);
        close(filelock);
    }
    const std::string& lock_filename = locations.lockFile();
    unlink(lock_filename.c_str());
    return 0;
}

/**
 * write_service_status
 * Write Service Status
 */
int write_service_status(
    const Locations&  locations, 
    int         name_len, 
    int         status){
    
    const std::string& lock_filename = locations.lockFile();
    int filelock = open(lock_filename.c_str(), O_WRONLY | O_TRUNC, 0600);
    if(-1 != filelock){
	ssize_t check_size = write(filelock, (char*)&status, sizeof(int));
	if(check_size <=0 )
	    fprintf(stdout,"Error occurred while writing lock file:%s\n", strerror(errno));
	fsync(filelock);
        close(filelock);
        return 0;
    }
    fprintf(stdout,"Error occurred while writing lock file:%s\n", strerror(errno));
    fflush(stdout);
    return -1;
}

/**
 * read_service_status
 * Read Service Status
 */
int read_service_status(const Locations& locations,int name_len, int * status){
    const std::string& lock_filename = locations.lockFile();
    int filelock = open(lock_filename.c_str(), O_RDONLY);
    int temp;
    if(-1 != filelock){
	read(filelock, (char*)&temp, sizeof(int));
        *status = temp;
	close(filelock);
        return 0;
    }
    fprintf(stdout,"Error occurred while reading lock file:%s\n", strerror(errno));
    fflush(stdout);
    return -1;
}
} // End namespace config
} // End namespace glite
