/**
 * Copyright (c) Members of the EGEE Collaboration. 2004.
 * See http://www.eu-egee.org/partners/ for details on the copyright holders.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *     http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 *  Common Functions for the GLite Data Config Service  module
 *
 *  Authors: Paolo Badino <paolo.badino@cern.ch>, Paolo Tedesco <paolo.tedesco@cern.ch>
 *  Version info: $Id: Functions.hxx,v 1.8 2008/01/24 08:43:44 ptedesco Exp $
 *  Release: $Name:  $
 */
#ifndef GLITE_DATA_CONFIG_FUNCTIONS_HXX_ 
#define GLITE_DATA_CONFIG_FUNCTIONS_HXX_

#include <string>

namespace glite {
namespace config{

class Locations;    
        
/**
 * pid_of_proc
 * Get pid of process, first looking at the file "/tmp/<name>.pid" and then at 
 * the /proc filesystem
 * @param locations [IN] location information for the process
 * @param name_len [IN] the meaningful length of the process name
 * @return the pid of the process or -1 in case of failure
 */
int pid_of_proc(const Locations& locations, int name_len);

/**
 * pid_of_running_proc
 * Get pid of a running process looking at /proc
 * @param name [IN] the name of the process to check
 * @param name_len [IN] the meaningful length of the process name
 * @return the pid of the process or -1 in case of failure
 */
int pid_of_running_proc(const Locations& locations, int name_len);  

/**
 * status_of_proc
 * Returns 0 is the process is running
 * @param name [IN] the name of the process to check
 * @param name_len [IN] the meaningful length of the process name
 * @return 0 is the process is running
 */
int status_of_proc(const Locations& locations, int name_len);

/**
 * check_pid
 * Check if pid is running
 * @param pid [IN] the id of the process to check
 * @return 0 if the process is running
 */
int check_pid(int pid);

/**
 * write_pidfile
 * Write the pid file for the given process
 * @param name [IN] the name of the process
 * @param name_len [IN] the meaningful length of the process name
 * @param location [IN] the location path used to create the pid file name
 * @param location_var [IN] the var location path used to create the pid file name
 * @param pid [IN] Optional, the id of the process. If 0 is specified, the pid 
 * is computed using pid_of_running_process
 * @return 0 if success
 */
int write_pidfile(const Locations& locations, int name_len, int pid = 0);
    
/**
 * unlink_pidfile
 * Delete the pid file for the given service
 * @param name [IN] the name of the process
 * @param name_len [IN] the meaningful length of the process name
 * @param location [IN] the location path used to create the pid file name
 * @param location_var [IN] the var location path used to create the pid file name
 * @return 0 if success
 */
int unlink_pidfile(const Locations& locations, int name_len);

/**
 * proc_acquire_lock
 * Acquire a lock for the process. This method will fail if another instance of 
 * the process is running
 * @param name [IN] the name of the process
 * @param name_len [IN] the meaningful length of the process name
 * @param location [IN] the location path used to create the lock file name
 * @param location_var [IN] the var location path used to create the lock file name
 * @return the lock file hanlde or -1 if failed
 */
int proc_acquire_lock(const Locations& locations);

/**
 * proc_release_lock
 * Release a lock for the process
 * @param name [IN] the name of the process
 * @param name_len [IN] the meaningful length of the process name
 * @param lock [IN] the file lock handle
 * @param location [IN] the location path used to create the lock file name
 * @param location_var [IN] the var location path used to create the lock file name
 * @return 0 if success
 */
int proc_release_lock(const Locations& locations, int name_len, int lock);

/**
 * Write the service Status to the Lock File
 * @param name [IN] the name of the process
 * @param name_len [IN] the meaningful length of the process name
 * @param location [IN] the location path used to create the lock file name
 * @param location_var [IN] the var location path used to create the lock file name
 * @param status [IN] the service status value
 * @return 0 if success
 */
int write_service_status(
    const Locations& locations, int name_len, int status);

/**
 * Read the Service Status from the Lock File
 * @param name [IN] the name of the process
 * @param name_len [IN] the meaningful length of the process name
 * @param location [IN] the location path used to create the lock file name
 * @param location_var [IN] the var location path used to create the lock file name
 * @param status [OUT] the service status value
 * @return 0 if success
 */
int read_service_status(const Locations& locations,int name_len, int * status);

} // End namespace config
} // End namespace glite

#endif //GLITE_DATA_CONFIG_FUNCTIONS_HXX_
