/**
 * Copyright (c) Members of the EGEE Collaboration. 2004.
 * See http://www.eu-egee.org/partners/ for details on the copyright holders.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *     http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * Locations.cxx
 *
 * Authors: Paolo Tedesco <paolo.tedesco@cern.ch>
 * Version info: $Id: Locations.cxx,v 1.3 2009/10/29 10:22:44 szamsu Exp $
 * Release: $Name:  $
 *
 */
#include "Locations.hxx"
#include <cstring>
#include <cstdlib>
#include <dirent.h>
#include <log4cpp/Category.hh>

#ifdef __READ_LOCATION_FILES
#include <fstream>
#endif

using namespace log4cpp;
using namespace glite::config;

namespace {

/*------------------------------------------------------------------------------
                                    Constants
------------------------------------------------------------------------------*/

// name of $GLITE_LOCATION environment variable
const char * GLITE_LOCATION_ENV = "GLITE_LOCATION";
// default value of $GLITE_LOCATION
const char * LOCATION_DEFAULT = "/opt/glite";

// name of $GLITE_LOCATION_VAR environment variable
const char * GLITE_LOCATION_VAR_ENV = "GLITE_LOCATION_VAR";
// default value of $GLITE_LOCATION_VAR
const char * VAR_LOCATION_DEFAULT = "/var/glite";

#ifdef __USE_LOCATION_USER
    // name of $GLITE_LOCATION_USER environment variable
    const char * GLITE_LOCATION_USER_ENV = "GLITE_LOCATION_USER";
    // default value of $GLITE_LOCATION_USER
    const char * USER_LOCATION_DEFAULT = "~/.glite";
#endif

#ifdef __READ_LOCATION_FILES
    // name of $HOME environment variable
    const char * HOME_ENV = "HOME";
    // configuration file in user's home directory 
    const char * HOME_CONF_FILE = ".glite.conf";
    // common configuration file path
    const char * COMMON_CONF_FILE = "/etc/glite.conf";
#endif

// extension for properties files
const char * PROPERTIES_EXT = ".properties.xml";
// extension for log properties files
const char * LOG_PROPERTIES_EXT = ".log-properties";
// extension for configuration directory
const char * DIRECTORY_EXT = ".d";

/*------------------------------------------------------------------------------
                                    Utilities
------------------------------------------------------------------------------*/

/**
 * dir_exists
 * @param dir_path [IN] directory path
 * @return true if directory exists, false otherwise
 */
bool dir_exists(const std::string& dir_path){
    bool ret = false;
    DIR * dir = opendir(dir_path.c_str());
    if (0 != dir){
        closedir(dir);
        ret = true;
    }
    return ret;
}

/**
 * create_file_path
 * get the path of a file combining path, name and extension
 * @param dir_path [IN]
 * @param file_name [IN]
 * @param file_ext [IN]
 * @return file path
 */
std::string create_file_path(
    const std::string& dir_path,
    const std::string& file_name,
    const std::string& file_ext){
    return dir_path + "/" + file_name + file_ext;
} 

/**
 * ValidationFunction
 * pointer to a function to determine if given file name is valid according to 
 * some criteria.
 * @param dir_path [IN] the directory full path
 * @param file_path [IN] the file full path
 * @return true if file name is valid, false otherwise
 */
typedef bool(*ValidationFunction)(const std::string& dir_path,const std::string& file_path);

/**
 * validate_file_name
 * @param dir_path [IN] directory path
 * @param file_name [IN] file name
 * @param file_ext [IN] file extension
 * @param validate [IN] validation function
 * @param file_path [OUT] the file path
 * @return true if file path is valid
 */
bool validate_file_name(
    const std::string& dir_path,
    const std::string& file_name,
    const std::string& file_ext,
    ValidationFunction validate,
    std::string& file_path){
    bool ret = false;
    file_path = create_file_path(dir_path,file_name,file_ext);
    ret = dir_exists(dir_path) && validate(dir_path,file_path);
    return ret;
}

/**
 * 
 * @param name [IN]
 * @param ext [IN]
 * @param path [IN]
 * @param alt_path [IN]
 * @param def_path [IN]
 * @param location [IN]
 * @param location_var [IN]
 * @param validate [IN]
 * @param file_path [IN]
 * @return true if successful, false otherwise
 */
bool get_valid_file_path(
    const char *        name,
    const char *        ext,
    const char *        path, 
    const char *        alt_path, 
    const char *        def_path, 
    const Locations&    locations,
    ValidationFunction  validate,
    std::string&        file_path){
        
    // check that the possible directories exist and validate file names; when
    // a valid file name is found in an existing directory, return
    std::string dir_name;
#ifdef __USE_LOCATION_USER
    const std::string& location_user = locations.location_user();
    if(!location_user.empty()) {
        dir_name = location_user + "/" + path;
        if (validate_file_name(dir_name,name,ext,validate,file_path)){
            return true;
        }
        dir_name = location_user + "/" + alt_path;
        if (validate_file_name(dir_name,name,ext,validate,file_path)){
            return true;
        }
    }
#endif        
    const std::string& location_var = locations.location_var();
    if(!location_var.empty()) {
        dir_name = "/" + std::string(path);
        if (validate_file_name(dir_name,name,ext,validate,file_path)){
            return true;
        }
        dir_name = "/" + std::string(alt_path);
        if (validate_file_name(dir_name,name,ext,validate,file_path)){
            return true;
        }
    }
    const std::string& location = locations.location();
    if(!location.empty()) {
        dir_name = "/" + std::string(path);
        if (validate_file_name(dir_name,name,ext,validate,file_path)){
            return true;
        }
        dir_name = "/" + std::string(alt_path);
        if (validate_file_name(dir_name,name,ext,validate,file_path)){
            return true;
        }
    }
    if (validate_file_name(def_path,name,ext,validate,file_path)){
        return true;
    }
    // could not find a valid file name in an existing directory
    file_path = "";
    return false;
}


/**
 * Check directory and file permissions for lock and pid files
 * @param dir_path [IN]
 * @param file_path [IN]
 * @return true if permissions on file and directory allow writing a file with
 * that name
 */
bool check_permissions(
    const std::string& dir_path,
    const std::string& file_path){
        
    // echo_debug("checking permissions on file %s",file_path.c_str());
    // does the file exist ?
    if(0 == access(file_path.c_str(), F_OK)){
        // file exists, check if it can be written
        if (0 == access(file_path.c_str(), W_OK)){
            // echo_debug("file %s exists and can be written -> ok",file_path.c_str());
            return true;
        } else {
            // echo_debug("file %s exists but cannot be written -> invalid",file_path.c_str());
            return false;
        }
    } else {
        // file does not exist; does the folder exist ?
        if(0 == access(dir_path.c_str(), F_OK)){
            // directory exists; check that we have execute and write permissions 
            // on it (can create a new file in the directory)
            if (0 == access(dir_path.c_str(),W_OK | X_OK)){
                // echo_debug("file %s does not exist but it can be created -> ok",file_path.c_str());
                return true;
            } else {
                // echo_debug("not enough privileges on parent directory of file %s -> invalid",file_path.c_str());
                return false;
            }
        } else {
            // directory doesn't exist
            // echo_debug("parent directory of file %s does not exist -> invalid",file_path.c_str());
            return false;
        }
    }    
}


/**
 * get the path of the lock file
 * @param name 
 * @param locations 
 * @param lock_file_path 
 * @return 
 */
bool get_lock_file_path(
    const char *        name,
    const Locations&    locations,
    std::string&        lock_file_path){
    // lock files constants
    static const char * LOCKFILE_DIR     = "lock";
    static const char * LOCKFILE_ALT_DIR = "tmp";
    static const char * LOCKFILE_DEF_DIR = "/tmp";
    static const char * LOCKFILE_EXT     = ".lock";
    // get lock file path
    return get_valid_file_path(name,LOCKFILE_EXT,LOCKFILE_DIR,LOCKFILE_ALT_DIR,LOCKFILE_DEF_DIR,locations,check_permissions,lock_file_path);
}

/**
 * get the path of the pid file
 * @param name 
 * @param locations 
 * @param pid_file_path 
 * @return 
 */
bool get_pid_file_path(
    const char *        name,
    const Locations&    locations,
    std::string&        pid_file_path){
    // pid files constants
    static const char * PIDFILE_DIR      = "/tmp";
    static const char * PIDFILE_ALT_DIR  = PIDFILE_DIR;
    static const char * PIDFILE_DEF_DIR  = "/tmp";
    static const char * PIDFILE_EXT      = ".pid";
    // get pid file path
    return get_valid_file_path(name,PIDFILE_EXT,PIDFILE_DIR,PIDFILE_ALT_DIR,PIDFILE_DEF_DIR,locations,check_permissions,pid_file_path);
}


}

/*------------------------------------------------------------------------------
                                Locations
------------------------------------------------------------------------------*/

/**
 * Constructor
 */
Locations::Locations() : 
    m_useAsPrefix(false),
    m_location(LOCATION_DEFAULT),
    m_locationVar(VAR_LOCATION_DEFAULT){
    
    const char * env_location      = getenv(GLITE_LOCATION_ENV);
    if (0 != env_location){
        m_location = env_location;
    }
    
    const char * env_location_var  = getenv(GLITE_LOCATION_VAR_ENV);
    if (0 != env_location_var){
        m_locationVar = env_location_var;
    }
    
#ifdef __USE_LOCATION_USER    
    const char * env_location_user = getenv(GLITE_LOCATION_USER_ENV);
    if (0 != env_location_user){
        m_locationUser = env_location_user;
    }
#endif

#ifdef __READ_LOCATION_FILES
    // Look At the $HOME/.glite.conf file
    const char * home = getenv(HOME_ENV);
    if(0 != home){
        std::string fname = (std::string)home + "/" + HOME_CONF_FILE;
        readLocationsFromFile(fname);
    }
        
    // Look At the /etc/glite.conf file
    std::string fname = COMMON_CONF_FILE;
    readLocationsFromFile(fname);
#endif

    if (m_location.empty()){
        m_location = LOCATION_DEFAULT;
    }
    if (m_locationVar.empty()){
        m_locationVar = VAR_LOCATION_DEFAULT;
    }
#ifdef __USE_LOCATION_USER    
    if (m_locationUser.empty()){
        m_locationUser = USER_LOCATION_DEFAULT;
    }
#endif

}


/**
 * Initialize the service information
 * @param serviceName [IN] the configuration namespace that should be used 
 * for the service
 */
void Locations::init_locations(
    const std::string&  serviceName){
    m_serviceName = serviceName;
    m_serviceInstanceName.clear();
    m_serviceDirName.clear();
    
    getPropertiesFilesPaths();
    getLockAndPidFilesPaths();
}

/**
 * Initialize the service information
 * @param serviceDirName [IN] the name of the configuration directory 
 * @param useAsPrefix [IN] use the configuration directory as prefix of 
 * the instance name
 * @param serviceInstanceName [IN] the name of the configuration instance. 
 * The configuration namespace for the instance is then computed depending 
 * on the value of the use_as_prefix parameter: if set to true the instance 
 * name will be <dir_name>-<name>, otherwise just name
 */
void Locations::init_locations(
    const std::string&  serviceInstanceName,
    const std::string&  serviceDirName,
    const bool&         useAsPrefix){

    m_serviceInstanceName = serviceInstanceName;
    m_serviceDirName = serviceDirName;
    m_useAsPrefix = useAsPrefix;
    if(m_useAsPrefix){
        m_serviceName = m_serviceDirName + "-" + m_serviceInstanceName;
    } else {
        m_serviceName = m_serviceInstanceName;
    }
    
    getPropertiesFilesPaths();
    getLockAndPidFilesPaths();
}

/**
 * get the paths of the configuration files (properties and log
 * properties)  
 */
void Locations::getPropertiesFilesPaths(){
    if (useDirAndInstanceName()){
        m_propertiesFilePath = getConfigFilePath(m_serviceDirName, m_serviceInstanceName + PROPERTIES_EXT);
        m_logPropertiesFilePath = getConfigFilePath(m_serviceDirName, m_serviceInstanceName + LOG_PROPERTIES_EXT);
    } else {
        m_propertiesFilePath = getConfigFilePath(m_serviceName + PROPERTIES_EXT);
        m_logPropertiesFilePath = getConfigFilePath(m_serviceName + LOG_PROPERTIES_EXT);
    }   
}

/**
 * get the lock and pid file paths for the service
 */ 
void Locations::getLockAndPidFilesPaths(){
    if (!get_pid_file_path(m_serviceName.c_str(),*this,m_pidFilePath)){
        throw ServiceConfigurationException("Could not determine a valid pid file path.");
    }
    if (!get_lock_file_path(m_serviceName.c_str(),*this,m_lockFilePath)){
        throw ServiceConfigurationException("Could not determine a valid lock file path.");
    }
}

/**
 * getConfigFilePath
 * Get File Path
 */
std::string Locations::getConfigFilePath(const std::string& dir, const std::string& filename){
    return getConfigFilePath(dir + DIRECTORY_EXT + (std::string)"/" + filename);
}

/**
 * getConfigFilePath
 * Get File Path
 */
std::string Locations::getConfigFilePath(const std::string& filename){

    std::string file_path = filename;
    
    bool found = false;

#ifdef __USE_LOCATION_USER        
    // Check GLITE_LOCATION_USER
    if(!m_locationUser.empty()){
        std::string pathname = (std::string)m_locationUser + "/etc/" + filename;
        // Check if file exists
        int fh = access(pathname.c_str(),R_OK);
        if(-1 != fh){
            // File Exists in GLITE_LOCATION_USER: Use that
            file_path = pathname;
            found = true;
        }
    } 
#endif    
    
    // Check GLITE_LOCATION_VAR
    if((false == found) && (0 != m_locationVar.length())){
        std::string pathname = (std::string)m_locationVar + "/etc/" + filename;
        // Check if file exists
        int fh = access(pathname.c_str(),R_OK);
        if(-1 != fh){
            // File Exists in GLITE_LOCATION_VAR: Use that
            file_path = pathname;
            found = true;
        }
    } 
    
    // Check GLITE_LOCATION
    if((false == found) && (0 != m_location.length())){
        std::string pathname = (std::string)m_location + "/etc/" + filename;
        // Check if file Exists
        int fh = access(pathname.c_str(),R_OK);
        if(-1 != fh){
            // File Exists in GLITE_LOCATION: Use that
            file_path = pathname;
            found = true;
        }
    }
   
    // Check GLITE_LOCATION_ETC
    if((false == found) && (0 != m_location.length())){
        std::string pathname =  "/etc/" + filename;
        // Check if file Exists
        int fh = access(pathname.c_str(),R_OK);
        if(-1 != fh){
            // File Exists in GLITE_LOCATION: Use that
            file_path = pathname;
            found = true;
        }
    }

 
    // Check current directory
    if(false == found){
        // Check if file Exists in local path
        int fh = access(filename.c_str(),R_OK);
        if(-1 != fh){
            // File Exists in local path
            file_path = filename;
            found = true;
        }
    }

    if(!found){
        std::stringstream reason;
        reason << "file <" << filename << "> not found in: ";
#ifdef __USE_LOCATION_USER        
        if(!m_locationUser.empty()){
            reason << m_locationUser << "/etc, ";
        }
#endif        
        if(!m_locationVar.empty()){
            reason << m_locationVar << "/etc, ";
        }
        if(!m_location.empty()){
            reason << m_location << "/etc, ";
        }
        char * pwd = getenv("PWD");
        if(0 != pwd){
            reason << pwd;
        } else {
            reason << "current working directory";
        }
        throw ServiceConfigurationException(reason.str());
    }
    
    return file_path;
}

#ifdef __READ_LOCATION_FILES

/**
 * readLocationsFromFile
 * @param fname 
 */
void Locations::readLocationsFromFile(
    const std::string&  fname){

    // Read the configuration File
    std::ifstream configFile(fname.c_str());
    if(configFile.is_open()){
        char line[1024]; 
        while(configFile.good()){
            configFile.getline(line,1024);

            // Check Separator
            char * pch = strchr(line,'=');
            if(0 == pch){
                continue;
            }
            *pch = '\0';

            // Get Property Name
            char * name = line;
                                                                                
            // Get Property Value
            char * value = pch + 1;
                                    
            // EatWhites on name
            size_t pos = strspn(name, " \t");
            char * name_begin = name + pos;
            pos = strcspn(name_begin, " \t");
            name_begin[pos] = 0;

            if((0 != strcmp(name_begin,GLITE_LOCATION_ENV)) &&
                (0 != strcmp(name_begin,GLITE_LOCATION_VAR_ENV))
#ifdef __USE_LOCATION_USER                
                 && (0 != strcmp(name_begin,GLITE_LOCATION_USER_ENV))
#endif
              ){
                continue;
            }
            // EatWhites on value
            pos = strspn(value, " \t");
            char * value_begin = value + pos;
                                                                            
            pos = 0;
            char * str = value_begin;
            do{
                // Check Black
                pos = strcspn(str, " \t");
                str += pos;
                if(pos == strlen(str) || 0 == pos){
                    break;
                }
                // Check Whites
                pos = strspn(value_begin, " \t");
                str += pos;
                if(pos == strlen(str) || 0 == pos){
                    break;
                }
            }while(pos == 0);
                                                                            
            value_begin[str - value_begin] = 0;

            // Lookup for the Right Properties
            if(m_location.empty() && (0 == strcmp(name_begin,GLITE_LOCATION_ENV))) {
                m_location = value_begin;
            } else if(m_locationVar.empty() && (0 == strcmp(name_begin,GLITE_LOCATION_VAR_ENV))) {
                m_locationVar = value_begin;
            }
#ifdef __USE_LOCATION_USER                
            else if(m_locationUser.empty() && (0 == strcmp(name_begin,GLITE_LOCATION_USER_ENV))) {
                m_locationUser = value_begin;
            }
#endif
        } // End while
    }
}  

#endif

