/**
 * Copyright (c) Members of the EGEE Collaboration. 2004.
 * See http://www.eu-egee.org/partners/ for details on the copyright holders.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *     http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 *  GLite Data Config Service : Utilities for locations of pid/lock and configuration files
 *
 *  Authors: Paolo Tedesco <paolo.tedesco@cern.ch>
 *  Version info: $Id: Locations.hxx,v 1.4 2008/06/12 14:23:53 ptedesco Exp $
 *  Release: $Name:  $
 */
#ifndef GLITE_DATA_CONFIG_LOCATIONS_HXX_ 
#define GLITE_DATA_CONFIG_LOCATIONS_HXX_

// define this to include $GLITE_LOCATION_USER in the search path
#define __USE_LOCATION_USER

// define this to read path values from configuration files
#define __READ_LOCATION_FILES

#include <string>
#include <glite/data/config/service/ServiceConfigurationException.hxx>

namespace glite {
namespace config{

/**
 * Holds information about service-related locations (GLITE_LOCATION etc) and
 * files (lock, pid and configuration files).
 */
class Locations{
    
public:
    
    /**
     * Constructor
     */
    Locations();

    /**
     * Initialize the service information
     * @param serviceName [IN] the configuration namespace that should be used 
     * for the service
     */
    void init_locations(
        const std::string&  serviceName) /* throw(ServiceConfigurationException) */;
        
    /**
     * Initialize the service information
     * @param serviceDirName [IN] the name of the configuration directory 
     * @param useAsPrefix [IN] use the configuration directory as prefix of 
     * the instance name
     * @param serviceInstanceName [IN] the name of the configuration instance. 
     * The configuration namespace for the instance is then computed depending 
     * on the value of the use_as_prefix parameter: if set to true the instance 
     * name will be <dir_name>-<name>, otherwise just name
     */
    void init_locations(
        const std::string&  serviceInstanceName,
        const std::string&  serviceDirName,
        const bool&         useAsPrefix) /* throw(ServiceConfigurationException) */;
    
    /**
     * serviceName
     * @return the service name
     */
    const std::string& serviceName() const {
        return m_serviceName;
    }
    
    /**
     * serviceInstanceName
     * @return the service instance name
     */
    const std::string& serviceInstanceName() const {
        return m_serviceInstanceName;
    }
    
    /**
     * serviceDirName
     * @return the service directory name
     */
    const std::string& serviceDirName() const {
        return m_serviceDirName;
    }
    
    /**
     * useAsPrefix
     * @return the use as prefix flag
     */
    bool useAsPrefix() const {
        return m_useAsPrefix;
    }
    
    /**
     * propertiesFile
     * @return the properties file path
     */
    const std::string& propertiesFile() const {
        return m_propertiesFilePath;
    }
    
    /**
     * logPropertiesFile
     * @return the log properties file path
     */
    const std::string& logPropertiesFile() const {
        return m_logPropertiesFilePath;
    }
    
    /**
     * lockFile 
     * @return the lock file path
     */
    const std::string& lockFile() const {
        return m_lockFilePath;
    }
    
    /**
     * pidFile
     * @return the pid file path
     */
    const std::string& pidFile() const {
        return m_pidFilePath;
    }
    
    /**
     * useDirAndInstanceName
     * @return true if directory and service instance name must be used.
     */
    bool useDirAndInstanceName() const {
        return (!m_serviceDirName.empty() && !m_serviceInstanceName.empty());
    }
    
    /**
     * location
     * @return the location path
     */
    const std::string& location() const {
        return m_location;
    }
    /**
     * location_var
     * @return the location_var path
     */
    const std::string& location_var() const {
        return m_locationVar;
    }

#ifdef __USE_LOCATION_USER        
    /**
     * location_user
     * @return the location_user path
     */
    const std::string& location_user() const{
        return m_locationUser;
    }
#endif

    /**
     * write debug messages with internal parameters
     */
    void debug_parameters();

private:

#ifdef __READ_LOCATION_FILES
    /**
     * readLocationsFromFile
     * @param fname [IN]
     */
    void readLocationsFromFile(const std::string&  fname);
#endif    
    
    /**
     * get the paths of the configuration files (properties and log
     * properties)  
     */
    void getPropertiesFilesPaths();
    
    /**
     * get the lock and pid file paths for the service
     */ 
    void getLockAndPidFilesPaths();
    
    /**
     * Get the path of a configuration file
     */
    std::string getConfigFilePath(const std::string& filename);
    
    /**
     * Get the path of a configuration file using the directory-instance name
     */
    std::string getConfigFilePath(const std::string& dir, const std::string& filename);

    // name of the service
    std::string m_serviceName;
    // service directory
    std::string m_serviceDirName;
    // name of service instance
    std::string m_serviceInstanceName;
    // use as prefix flag
    bool m_useAsPrefix;
    
    // path of the lock file
    std::string m_lockFilePath;
    // path of the pid file
    std::string m_pidFilePath;
    
    // location path
    std::string m_location;
    // location_var path
    std::string m_locationVar;
#ifdef __USE_LOCATION_USER        
    // location_user path
    std::string m_locationUser;
#endif    
    
    // path of the properties file
    std::string m_propertiesFilePath;
    // path of the log properties file
    std::string m_logPropertiesFilePath;

};    
    
} // namespace config
} // namespace glite

#endif
