/*
 * Copyright (c) Members of the EGEE Collaboration. 2004.
 * See http://www.eu-egee.org/partners/ for details on the copyright holders.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *     http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * main file for the GLite Data Conf Service  module
 *
 * Authors: Paolo Badino <paolo.badino@cern.ch>, Paolo Tedesco <paolo.tedesco@cern.ch>
 * Version info: $Id: Service.cxx,v 1.27 2008/06/12 14:23:53 ptedesco Exp $
 * Release: $Name:  $
 *
 */
#include "glite/data/config/service/ServiceConfigurator.hxx"
#include "Functions.hxx"
#include "Echo.hxx"
#include "Locations.hxx"

#include <stdarg.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <errno.h>
#include <unistd.h>
#include <string.h>
#include <signal.h>
#include <sys/wait.h>
#include <string>
#include <sstream>
#include <getopt.h>


   
using namespace glite::config;
namespace {

const char * PROGNAME   = "glite-data-config-service";
const char * VERSION    = CONFIG_SERVICE_VERSION;
const char * AUTHOR     = "(C) Paolo Badino - paolo.badino@cern.ch";

const int SERVICE_STOPPED    = 0;
const int SERVICE_STARTING   = 1;
const int SERVICE_STARTED    = 2;
const int SERVICE_STOPPING   = 3;
const int SERVICE_FAILED     = 4;
}
/*
 * Declare signal handler
 */
static int global_name_len = 0; 
static Locations global_locations; 
void sighandler(int signum); 
 
void empty_handler(int signum){
}

/**
 * run_service
 * run the service application
 * @param name          [IN] base name for the service 
 * @param dirname       [IN] 
 * @param is_prefix     [IN] if true, dirname is considered as a prefix for service
 * name, and service name is formed as <dirname>-<name>
 * @param name_len      [IN] 
 * @param as_service    [IN]
 * @return 
 */
int run_service(
    const Locations&  locations, 
    int         name_len, 
    bool        as_service) {

    // signals to block
    sigset_t signal_mask;
    sigemptyset(&signal_mask);
    sigaddset(&signal_mask, SIGINT);
    sigaddset(&signal_mask, SIGTERM);
    sigaddset(&signal_mask, SIGHUP);
    
    // The process Blocks all signals except SIGINT, SIGTERM and SIGUP
    sigprocmask(SIG_SETMASK,&signal_mask,NULL);

    // Block The signals until the service is started
    sigprocmask(SIG_BLOCK,&signal_mask,NULL);
    // Set Signal Handlers
    signal (SIGTERM, empty_handler);
    signal (SIGINT,  empty_handler);
    signal (SIGHUP,  empty_handler);
    

    // Instantiate the ServiceConfigurator
    ServiceConfigurator * serviceConfigurator = ServiceConfigurator::instance();

    int result = serviceConfigurator->init(&locations);
    if(0 != result){
        if(true == as_service){
            //write_service_status(service_name.c_str(),name_len,location.c_str(),location_var.c_str(),SERVICE_FAILED);
            write_service_status(locations,name_len,SERVICE_FAILED);
        }
        ServiceConfigurator::dispose();
        return EXIT_FAILURE;
    }
    // Start The Service
    result = serviceConfigurator->start();
    if(0 != result){
        if(true == as_service){
            write_service_status(locations,name_len,SERVICE_FAILED);
        }
        ServiceConfigurator::dispose();
        return EXIT_FAILURE;
    }

    // Update Service Status
    if(true == as_service){
        write_service_status(locations,name_len,SERVICE_STARTED);
    }

    // Do Main Loop    
    int doLoop = 1;
    while(doLoop == 1){
        
        // Enable Signal Handling
        sigprocmask(SIG_UNBLOCK,&signal_mask,NULL);

        // Wait until a signal is received
        int sig_caught;
        sigwait(&signal_mask, &sig_caught);
	
        
        // Block Signal Handling
        sigprocmask(SIG_BLOCK,&signal_mask,NULL);
        // Check Signal
        switch (sig_caught){
            case SIGINT:
            case SIGTERM:
                // Stop the Service
                doLoop = 0;
            break;
            case SIGHUP:
                // Reconfigure the Service
                result = serviceConfigurator->reconfig();
                if(0 != result){
                    // Exit With Error
                    doLoop = -1;
                } else {
                    // Continue Loop
                    doLoop = 1;
                }
            break;
        }
    }

    // Update Service Status
    if(true == as_service){
        write_service_status(locations,name_len,SERVICE_STOPPING);
    }

    // Stop The Service
    result = serviceConfigurator->stop();
 
    // Finalize The Service
    result = serviceConfigurator->fini();
    if((doLoop < 0) || (0 != result)){
        ServiceConfigurator::dispose();
        return EXIT_FAILURE;
    }
    
    // Delete the Service Configurator
    ServiceConfigurator::dispose();
    
    return 0;
}

/**
 * start_service
 * start the service application in a separate process
 * @param name      [IN]
 * @param dirname   [IN]
 * @param is_prefix [IN]
 * @param name_len  [IN]
 * @return 
 */
int start_service(
    const Locations&  locations, 
    int         name_len){

    std::string action = (std::string)"Starting Service " + locations.serviceName();
    echo_action(action.c_str());  
    
    // Process ID and Session ID
    pid_t pid, sid;
        
    // Acquire lock
    int lock = proc_acquire_lock(locations);
    if(-1 == lock){
        echo_failure();  
        echo_message("An instance of the service is already running");
        return -1;
    }
    // Check if there is already a process running
    pid = pid_of_running_proc(locations,name_len);
    if(-1 != pid){
        echo_failure();  
        echo_message("An instance of the service is already running but the lock file has been deleted");
        proc_release_lock(locations,name_len,lock);
        return -1;
    }
    
    // Set Service Status
    write_service_status(locations,name_len,SERVICE_STARTING);
    // Fork off the parent process
    pid = fork();
    if (pid < 0) {
        echo_failure();  
        echo_message("Cannot fork child process");
        proc_release_lock(locations,name_len,lock);
        return -1;
    }
    // If Pid is good PID, exit the parent process
    if (pid > 0) {
        // Sleep For a While
        sleep(1);
        // Get Process Status
        int result = check_pid(pid);
        if(0 != result){
            echo_failure(); 
            std::stringstream reason;
            reason << "Service failed to start (check_pid returned " << result << ") errno:" << strerror(errno) << "";
            echo_message(reason.str().c_str());
        } else {
            int svc_status = SERVICE_STARTING;
            int count = 0;
            // Check If the Service is Really Started
            do{
                sleep(1);
                count++;
                read_service_status(locations,name_len,&svc_status);
            } while((svc_status == SERVICE_STARTING) && (count < 60)); 
            switch(svc_status){
                case SERVICE_STARTED:
                    echo_success();
                break;
                case SERVICE_STARTING:
                    echo_warning();
                    echo_message("Service still starting after 60 seconds");
                break;
                default:
                    echo_failure();  
                    std::stringstream reason;
                    reason << "Service failed to start (svc_status is " << svc_status << ")";
                    echo_message(reason.str().c_str());
                break;
            }
        }
        return result;
    }
    
    // Change The Service Name
    ServiceConfigurator::setProcName(locations.serviceName().c_str());
        
    // Write pid in pidfile
    write_pidfile(locations,name_len,getpid());
    
    
    // Create a new SID for the child process
    sid = setsid();
    if (sid < 0) {
        echo_message("Cannot get a session id");
        proc_release_lock(locations,name_len,lock);
        unlink_pidfile(locations,name_len);
        return -1 ;
    }
    
    // Close out the standard file descriptors
    close(STDIN_FILENO);
    close(STDOUT_FILENO);
    close(STDERR_FILENO);
    
    struct sigaction SignalAction;
    SignalAction.sa_flags = 0;
    SignalAction.sa_handler = sighandler;
    sigemptyset(&SignalAction.sa_mask);
    sigaction(SIGSEGV, &SignalAction, NULL);
    
    // Run the service
    int result = run_service(locations,name_len,true);
    if(0 != result) {
        echo_message("Failed To Run Service");
    }
    
    // Release Synch Flag
    write_service_status(locations,name_len,SERVICE_STOPPED);
    
    // Release lock
    proc_release_lock(locations,name_len,lock);

    // Delete pidfile
    unlink_pidfile(locations,name_len);
        
    return result;
}

/**
 * stop_service
 * Stop the running service application
 * @param name 
 * @param dirname 
 * @param is_prefix 
 * @param name_len 
 * @param wait_timeout 
 * @param warn_if_not_exists 
 * @return 
 */
int stop_service(
    const Locations&    locations, 
    int                 name_len, 
    int                 wait_timeout = 60, 
    bool                warn_if_not_exists = false){
  
    const std::string& service_name = locations.serviceName();
    bool is_warning = false;
    int result = -1;
    std::string msg;
    std::string action = (std::string)"Stopping Service " + service_name;
    echo_action(action.c_str());  
    int pid = pid_of_proc(locations,name_len);
    if(pid > 0){
        result = kill(pid, SIGTERM);
        if(0 == result) {
            // Wait that the file is stopped
            int count=0;
            while((0 == result) && (count < wait_timeout)){
                result = check_pid(pid);
                sleep(1);
                count++;
            }
            if(count >= wait_timeout){
                std::stringstream streamstr;
                streamstr << "Failed to Stop Service " << service_name << " in " << wait_timeout << " seconds. Try again or kill the service";
                msg = streamstr.str();
                result = -1;
            } else {
                result = 0;
            }
        } else {
            msg = "Failed To send TERM signal";
        }
    } else {
        // Release lock
        proc_release_lock(locations,name_len,-1);

        // Delete pidfile
        unlink_pidfile(locations,name_len);

        msg = "No process running";
        is_warning = warn_if_not_exists;
        result = 1;
    }
    if(0 == result){
        echo_success();  
    } else {
        if(is_warning == true){
            echo_warning();
        } else {
            echo_failure(); 
        }
        echo_message(msg.c_str());
    }
    return result;
}

/**
 * kill_service
 * Kill the running service application
 * @param name 
 * @param dirname 
 * @param is_prefix 
 * @param name_len 
 * @return 
 */
int kill_service(
    const Locations& locations, 
    int                 name_len){
        
    const std::string& service_name = locations.serviceName();
    int result = -1;
    std::string action = (std::string)"Killing Service " + service_name;
    echo_action(action.c_str());  
    int pid = pid_of_proc(locations,name_len);
    if(pid > 0){
        result = kill(pid, SIGKILL);
        // Sleep for a while 
        sleep(1);
    }
    if(0 == result){
        // Release lock
        proc_release_lock(locations,name_len,-1);
        // Delete pidfile
        unlink_pidfile(locations,name_len);
        echo_success();
    } else {
        echo_failure(); 
    }
    return 0;
}


/**
 * service_status
 * Display the Service status
 * @param name 
 * @param dirname 
 * @param is_prefix 
 * @param name_len 
 * @return 
 */
int service_status(
    const Locations&  locations, 
    int                 name_len){
  
    const std::string& service_name = locations.serviceName();
    
    int status_result = -1;
    std::string msg = (std::string)"Service " + service_name + " status is ";
    int result = status_of_proc(locations,name_len);
    if(0 == result){
        int svc_status = SERVICE_STARTED;
        // Get Service Status
        read_service_status(locations,name_len,&svc_status);
        switch(svc_status){
            case SERVICE_STARTING:
                msg += "starting";
                status_result = 0;
            break;
            case SERVICE_STARTED:
                msg += "running";
                status_result = 0;
            break;
            case SERVICE_STOPPING:
                msg += "stopping";
                status_result = 0;
            break;
            case SERVICE_STOPPED:
                msg += "stopped";
                status_result = 1;
            break;
            default:
                msg += "failed";
                status_result = -1;
            break;
        }
    } else {
        msg += "stopped";
        status_result = 1;
    }
    echo_message(msg.c_str());   
    return status_result;
}

/**
 * reconfig_service
 * reconfigure a service
 * @param name 
 * @param dirname 
 * @param is_prefix 
 * @param name_len 
 * @return 
 */
int reconfig_service(
    const Locations&    locations, 
    int                 name_len){
  
    const std::string& service_name = locations.serviceName();
    
    int result = -1;
    std::string action = (std::string)"Reconfiguring Service " + service_name;
    echo_action(action.c_str());  
    std::string msg;
    int pid = pid_of_proc(locations,name_len);
    if(pid > 0){
        // Send SIGHUP Signal 
        result = kill(pid, SIGHUP);
        if(0 == result){
            // Sleep For a While
            sleep(3);
            // Get Process Status
            result = status_of_proc(locations,name_len);
            if(0 != result){
                msg = "Service Stopped during reconfiguration";
            }
        }
    } else {
        msg = "Service is not Running";
    }
    if(0 == result){
        echo_success();  
    } else {
        echo_failure();  
        echo_message(msg.c_str());   
    }
    return result;
}

/*
 * print_usage
 * 
 * Print Usage
 */
void print_usage(){
    fprintf(stderr,"\n");
    fprintf(stderr,"<%s> Version %s by %s\n",PROGNAME,VERSION,AUTHOR);
    fprintf(stderr,"\nUsage:\n");
    fprintf(stderr,"%s [-Q] [-d <dirname> | -p <name>] <name> <action>\n",PROGNAME);
    fprintf(stderr,"where:\n");
    fprintf(stderr,"\t-Q | --quiet:     Don't prompt on stdout/err\n");
    fprintf(stderr,"\t-v | --verbose:   Print debug messages\n");
    fprintf(stderr,"\t-d | --directory: use a directory configuration. In this case the <name>\n"\
                   "\t                  parameter is interpreted as configuration instance\n"\
                   "\t                  name and the Service application name will be <name>\n");
    fprintf(stderr,"\t-p | --prefix:    act as the directory configuration, but in this case the\n"\
                   "\t                  Service application name will be <prefix>-<name>\n");
    fprintf(stderr,"\tname:             the Service Application name\n");
    fprintf(stderr,"\taction:           The action to be performed\n");
    fprintf(stderr,"\t        start:    start a service in a separate process\n");
    fprintf(stderr,"\t        stop:     stop a service running in a separate process\n");
    fprintf(stderr,"\t        kill:     kill a service running in a separate process. Resource will not be released !\n");
    fprintf(stderr,"\t        restart:  restart a service\n");
    fprintf(stderr,"\t        reconfig: reconfigure a service without stopping it\n");
    fprintf(stderr,"\t        status:   print the service status\n");
    fprintf(stderr,"\t        debug:    run the service in the current process\n");
    fprintf(stderr,"\n");
}

/*
 * string convertInt
 * 
 * Convert string to int
 */
static std::string convertInt(int number)
{
   //create a stringstream
   std::stringstream ss;
   //add number to the stream
   ss << number;
   return ss.str();
}


/*
 * sighandler
 * 
 * Catch SIGSEGV signal and cleanup resources (if it is possible) before the process stack gets totally corrupted
 */
void sighandler(int signum)
{
	int result = 0;
	int pid = pid_of_proc(global_locations,global_name_len);
	FILE *ptr;
	ServiceConfigurator * serviceConfigurator = ServiceConfigurator::instance();
	
	//get a image of the process in memory for debugging (stack trace)-if it is possible
	std::string cmd = "pstack ";
	cmd.append( convertInt(pid) );
	cmd.append(" > /tmp/agent-stackframes.log");
	const char * pStr = cmd.c_str();
     
	if ((ptr = popen(pStr, "w")) != NULL)
		(void) pclose(ptr);
		
	echo_message("Segmantation fault, now releasing all the resources and exiting.");
	echo_message("Check /tmp/agent-stackframes.log for stack frame.");
	//now release all the resources and stop the service
	if (0 != serviceConfigurator)
	{
		result = serviceConfigurator->stop();
 
		// Finalize The Service
		result = serviceConfigurator->fini();
		if((0 != result)){
			ServiceConfigurator::dispose();       
		}
	}
			 
	//release lock
	proc_release_lock(global_locations,global_name_len,-1);
	// Delete pidfile
	unlink_pidfile(global_locations,global_name_len);
	//abort, no way to recover, force coredump generation
	abort();
}


/*
 * main
 * 
 * main function
 */
int main(int argc, char * argv[]) {

    int name_len = ServiceConfigurator::initProcNameVars(argc, argv);
    global_name_len = name_len;     

    bool enable_echo = true;
    bool enable_debug = false;
    std::string name;
    std::string dir_name;
    bool is_prefix = false;
    bool is_dir_already_set = false;
    std::string action;
    
    // options descriptor
    static struct option longopts[] = {
        {"quiet",     no_argument,       NULL, 'Q'},
        {"verbose",   no_argument,       NULL, 'v'},
        {"directory", required_argument, NULL, 'd'},
        {"prefix" ,   required_argument, NULL, 'p'},
        {NULL,        0,                 NULL, 0  }
     };

     int ch;
     while((ch = getopt_long(argc, argv, "Qvd:p:", longopts, NULL)) != -1){
         switch (ch) {
            case 'Q':
                enable_echo = false;
            break;
            case 'v':
                enable_debug = true;
            break;
            case 'd':
                if(true == is_dir_already_set){
                    print_usage();
                    return -1;
                }
                is_dir_already_set = true;
                if(0 != optarg){
                    dir_name = optarg;
                    is_prefix = false;
                } else {
                    print_usage();
                    return -1;
                }
                break;
            case 'p':
                if(true == is_dir_already_set){
                    print_usage();
                    return -1;
                }
                is_dir_already_set = true;
                if(0 != optarg){
                    dir_name = optarg;
                    is_prefix = true;
                } else {
                    print_usage();
                    return -1;
                }
                break;
             default:
                print_usage();
                return -1;
        }// End switch
    } // end while
    
    if((argc - optind) < 2) {
        print_usage();
        return -1; 
    }
    argv += optind;
    // Get instance name
    if(0 == argv[0]){
        print_usage();
        return -1; 
    }
    name = argv[0];
    // Get action
    if(0 == argv[1]){
        print_usage();
        return -1; 
    }
    action = argv[1];

    // enable echo/debug messages
    echo_enabled(enable_echo);
    debug_enabled(enable_debug);
    
    // Check The Action
    Locations locations;
    try{
        if (dir_name.empty()){
            locations.init_locations(name);
        } else{
            locations.init_locations(name,dir_name,is_prefix);
        }
    } catch (const ServiceConfigurationException& exc) {
        echo_message(exc.what());
        return -1; 
    } catch (const std::exception& exc) {
        echo_message(exc.what());
        return -1; 
    }
    
    global_locations = locations;
    
    if (enable_debug){
        echo_debug("serviceName         = %s",locations.serviceName().c_str());
        echo_debug("serviceInstanceName = %s",locations.serviceInstanceName().c_str());
        echo_debug("serviceDirName      = %s",locations.serviceDirName().c_str());
        echo_debug("propertiesFile      = %s",locations.propertiesFile().c_str());
        echo_debug("logPropertiesFile   = %s",locations.logPropertiesFile().c_str());
        echo_debug("lockFile            = %s",locations.lockFile().c_str());
        echo_debug("pidFile             = %s",locations.pidFile().c_str());
        echo_debug("location            = %s",locations.location().c_str());
        echo_debug("location_var        = %s",locations.location_var().c_str());
#ifdef __USE_LOCATION_USER
        echo_debug("location_user       = %s",locations.location_user().c_str());
#endif
    }
        
    int result;
    if(action == "start"){
        // Start the Service
        result = start_service(locations, name_len);
    } else if(action == "stop"){
        result = stop_service(locations, name_len);
    } else if(action == "kill"){
        result = kill_service(locations, name_len);
    } else if(action == "restart"){
        result = stop_service(locations, name_len, 120, true);
        if(-1 == result){
            result = kill_service(locations, name_len);
        }
        result = start_service(locations, name_len);
    } else if(action == "debug"){
        result = run_service(locations, name_len, false);
    } else if(action == "status"){
        result = service_status(locations, name_len);
    } else if(action == "reconfig"){
        result = reconfig_service(locations, name_len);
    } else {
        print_usage();
        return -1; 
    }
    

    return result;
}
