/**
 * Copyright (c) Members of the EGEE Collaboration. 2004.
 * See http://www.eu-egee.org/partners/ for details on the copyright holders.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *     http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * ServiceConfiguration.cxx
 *
 * Authors: Paolo Badino <paolo.badino@cern.ch>
 *          Paolo Tedesco <paolo.tedesco@cern.ch>
 * Version info: $Id: ServiceConfigurator.cxx,v 1.28 2009/12/03 17:18:01 molnarzs Exp $
 * Release: $Name:  $
 *
 */
#include "glite/data/config/service/ServiceConfigurator.hxx"
#include "glite/data/config/service/ParamValue.hxx"

#include "Locations.hxx"

#include <unistd.h>
#include <dlfcn.h>
#include <stdexcept>
#include <fstream>
#include <time.h>
#include <sstream>
#include <cstring>
#include <syslog.h>

#include <log4cpp/Category.hh>
#include <log4cpp/PropertyConfigurator.hh>
#include <log4cpp/PatternLayout.hh>

#include "log-message.hxx"

#include <glite/data/config/service/SysLogAuthpriv.h>
#include <glite/data/config/service/SysLogDaemon.h>
#include <glite/data/config/service/SysLogInfo.h>

using namespace log4cpp;
using namespace glite::config;

namespace{
    
const char * LEGAL_FILENAME_CHARS    = 
    "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789.-_+@&%,!$*()/\\";
    
const char * CREATE_COMPONENT_API    = "create_glite_component";
const char * DESTROY_COMPONENT_API   = "destroy_glite_component";


} // end global namespace

#define LOG_ERROR_MESSAGE(a) {std::stringstream msg;msg << a;log_error_message(msg.str().c_str());}
#define LOG_DEBUG_MESSAGE(a) {std::stringstream msg;msg << a;log_debug_message(msg.str().c_str());}

/*
 * Static Variables
 */
ServiceConfigurator * ServiceConfigurator::s_instance = 0;
char **               ServiceConfigurator::s_argv     = 0;
int                   ServiceConfigurator::s_nameLen  = 0;

/**
 * instance
 * Return singleton instance
 */
ServiceConfigurator * ServiceConfigurator::instance(){
    if(0 == s_instance){
        s_instance = new ServiceConfigurator();
    }
    
    return s_instance;
}

/**
 * dispose
 * Delete singleton instance
 */
void ServiceConfigurator::dispose(){
    if(0 != s_instance){
        // Check Status
        delete s_instance;
        s_instance = 0;
    }
}

/**
 * ServiceConfigurator
 * Constructor
 */
ServiceConfigurator::ServiceConfigurator():
    m_status(SC_STATUS_IDLE),
    m_logger(Category::getInstance("ServiceConfigurator")),
    m_locations(0){
}
    
/**
 * ~ServiceConfigurator
 * Destructor
 */
ServiceConfigurator::~ServiceConfigurator(){
    if(SC_STATUS_STARTED == m_status){
        this->stop();
    }
    if((SC_STATUS_INITIALIZED == m_status) || (SC_STATUS_FAILED == m_status)){
        this->fini();
    }
}

/**
 * init
 */
int ServiceConfigurator::init(const Locations * locations){
    
    int result = SC_FAILED;
    if (SC_STATUS_INITIALIZED == m_status){
        result = SC_ALREADY_INIT;
    } else if(SC_STATUS_IDLE == m_status) {
        
        m_locations = locations;
        // Store names
        m_name = m_locations->serviceName();
        m_instanceName = m_locations->serviceInstanceName();
        m_dirName = m_locations->serviceDirName();

        if (m_locations->useDirAndInstanceName()){
            if(false == validateName(m_dirName.c_str())){
                LOG_ERROR_MESSAGE("Service initialization failed: invalid character(s) in the directory name " << m_dirName);
                return SC_FAILED;
            }
            if(m_dirName.at(m_dirName.length() - 1) == '/'){
                LOG_ERROR_MESSAGE("Service initialization failed: directory name should not terminate with a slash: " << m_dirName);
                return SC_FAILED;
            }
            if(false == validateName(m_instanceName.c_str())){
                LOG_ERROR_MESSAGE("Service initialization failed: invalid character(s) in the instance name " << m_instanceName);
                return SC_FAILED;
            }
        } else {
            // validate name
            if(false == validateName(m_name.c_str())){
                LOG_ERROR_MESSAGE("Service initialization failed: invalid character(s) in the instance name: " << m_name);
                return SC_FAILED;
            }
        }
        
        // Init Logger
        result = initLogger();
        if(0 == result){
            // Load Components
            result = loadComponents();
            if(0 == result){
                // Init Components
                result = initComponents();
            }
        }
        if(0 == result){
            result = SC_SUCCESS;
            m_status = SC_STATUS_INITIALIZED;
        } else {
            result = SC_FAILED;
            m_status = SC_STATUS_FAILED;
        }
    } else if(SC_STATUS_STARTED == m_status){
        result = SC_ALREADY_STARTED;
    }
    return result;
    
}

/*
 * initComponents
 * 
 * Initailize the Components
 */
int ServiceConfigurator::initComponents(){
    int result = 0;
    
    m_logger.debug("Service %s - Initialize Components - Phase 1",m_name.c_str());
    
    ComponentList::const_iterator it;
    for(it = m_componentList.begin(); it != m_componentList.end(); it++){
        ComponentConfiguration * component = (*it)->component;
        // Initialize The Component
        try {
            result = component->init((*it)->init_params);
        } catch (const ServiceConfigurationException& exc) {
            m_logger.error("Service configuration exception during initialization of component %s: %s",
                component->getName(),
                exc.what());
            result = -1;
        }
        if(0 != result){
            m_logger.error("Service %s - Failed To Initialize Component %s",m_name.c_str(), component->getName());
            break;
        }
    }
    if(0 != result){
        return result;
    }

    m_logger.debug("Service %s - Initialize Components - Phase 2",m_name.c_str());

    for(it = m_componentList.begin(); it != m_componentList.end(); it++){
        ComponentConfiguration * component = (*it)->component;
        // Configure The Component
        try {
            result = component->config((*it)->config_params);
        } catch (const ServiceConfigurationException& exc) {
            m_logger.error("Service configuration exception during initialization of component %s: %s",
                component->getName(),
                exc.what());
            result = -1;
        }
        if(0 != result){
            m_logger.error("Service %s - Failed To Configure Component %s",m_name.c_str(), component->getName());
            break;
        }
    }
    if(0 == result){
        m_logger.debug("Service %s - Components Initialized",m_name.c_str());
    }
    
    return result;
}

/*
 * start
 * 
 * Start the Service
 */
int ServiceConfigurator::start(){
    int result = SC_FAILED;
    if(SC_STATUS_STARTED == m_status){
        result = SC_ALREADY_STARTED;
    } else if(SC_STATUS_IDLE == m_status) {
        result = SC_NOT_INIT;
    } else if(SC_STATUS_INITIALIZED == m_status) {
    
        m_logger.debug("Service %s - Starting Services",m_name.c_str());
        
        ComponentList::const_iterator it;
        for(it = m_componentList.begin(); it != m_componentList.end(); it++){
            ComponentConfiguration * component = (*it)->component;
            // Start the Component
            result = component->start();
            if(0 != result){
                m_logger.error("Service %s - Failed To Start Component %s",m_name.c_str(), component->getName());
                break;
            }
        }
        if(0 == result){
            m_logger.debug("Service %s - Services Started",m_name.c_str());
            SysLogDaemon().event(EventNames::ServiceStart).success(true).notice();
            m_status = SC_STATUS_STARTED;
            result = SC_SUCCESS;
        } else {
            SysLogDaemon().event(EventNames::ServiceStart).success(false).error();
            m_status = SC_STATUS_FAILED;
            result = SC_FAILED;
        }
    }
    return result;
}


/*
 * stop
 * 
 * Stop the Service
 */
int ServiceConfigurator::stop(){
    int result = SC_FAILED;
    if(SC_STATUS_INITIALIZED == m_status){
        result = SC_NOT_STARTED;
    } else if(SC_STATUS_IDLE == m_status) {
        result = SC_NOT_INIT;
    } else if(SC_STATUS_STARTED == m_status) {
    
        m_logger.debug("Service %s - Stopping Services",m_name.c_str());
    
        ComponentList::reverse_iterator rit;
        for(rit = m_componentList.rbegin(); rit != m_componentList.rend(); rit++){
            ComponentConfiguration * component = (*rit)->component;
            // Stop The Component
            result = component->stop();
            if(0 != result){
                m_logger.warn("Service %s - Failed To Stop Component %s",m_name.c_str(), component->getName());
                //Do Not Break
                //break;
            }
        }
        if(0 == result){
            m_logger.debug("Service %s - Services Stopped",m_name.c_str());
            SysLogDaemon().event(EventNames::ServiceStop).success(true).notice();
            m_status = SC_STATUS_INITIALIZED;
            result = SC_SUCCESS;
        } else {
            m_logger.debug("Service %s - Services Stopped With Errors",m_name.c_str());
            SysLogDaemon().event(EventNames::ServiceStop).success(false).error();
            m_status = SC_STATUS_INITIALIZED;
            result = SC_FAILED;
        }
    } 
    return result;
}

/*
 * reconfig
 * 
 * Reconfigure the Service
 */
int ServiceConfigurator::reconfig(){
    
    m_logger.debug("Service %s - Reconfiguring the Service",m_name.c_str());
    
    int result = this->stop();
    if(0 != result){
        m_logger.warn("Service %s - Cannot Reconfigure: Failed To Stop the Service",m_name.c_str());
        SysLogDaemon().event(EventNames::ServiceReconfigure).success(false).msg("Failed to stop the service").error();
        return result;
    }
    
    // Clear Config Params
    ComponentList::const_iterator it;
    for(it = m_componentList.begin(); it != m_componentList.end(); it++){
        ComponentConfiguration::Params& params = (*it)->config_params;
        ComponentConfiguration::Params::iterator pit;
        for(pit = params.begin(); pit != params.end(); pit++){
             delete (*pit).second;
        }
        params.clear();
    }
    
    result = initLogger();
    if(0 != result){
        m_logger.warn("Service %s - Cannot Reconfigure: Failed To Reconfigure Logger",m_name.c_str());
        SysLogDaemon().event(EventNames::ServiceReconfigure).success(false).msg("Failed to reconfigure the logger").error();
        m_status = SC_STATUS_FAILED;
        return SC_FAILED;
    }
    
    m_logger.debug("Service %s - Reloading Configuration",m_name.c_str());

    result = reloadConfiguration();
    if(0 != result){
        m_logger.warn("Service %s - Cannot Reconfigure: Failed To Reload Configuration",m_name.c_str());
        SysLogDaemon().event(EventNames::ServiceReconfigure).success(false).msg("Failed to reload configuration").error();
        m_status = SC_STATUS_FAILED;
        return SC_FAILED;
    }

    m_logger.debug("Service %s - Reconfiguring Components",m_name.c_str());

    for(it = m_componentList.begin(); it != m_componentList.end(); it++){
        ComponentConfiguration * component = (*it)->component;
        // Configure The Component
        try{
            result = component->config((*it)->config_params);
        } catch (const ServiceConfigurationException& exc) {
            m_logger.error("Service configuration exception during initialization of component %s: %s",
                component->getName(),
                exc.what());
            result = -1;
        }

        if(0 != result){
            m_logger.error("Service %s - Failed To Reconfigure Component %s",m_name.c_str(), component->getName());
            break;
        }
    }
    if(0 != result){
        m_logger.debug("Service %s - Cannot Reconfigure: Failed To Reconfigure Components",m_name.c_str());
        SysLogDaemon().event(EventNames::ServiceReconfigure).success(false).msg("Failed to reconfigure components").error();
        m_status = SC_STATUS_FAILED;
        return SC_FAILED;
    }
    
    result = this->start();
    if(0 != result){
        m_logger.warn("Service %s - Cannot Reconfigure: Failed To Start the Service",m_name.c_str());
        SysLogDaemon().event(EventNames::ServiceReconfigure).success(false).msg("Failed to start the service").error();
        return result;
    }
    
    SysLogDaemon().event(EventNames::ServiceReconfigure).success(true).notice();
    m_logger.debug("Service %s - Reconfiguration Succeeded",m_name.c_str());
    return 0;
}

/*
 * fini
 * 
 * Finalize the Service
 */
int ServiceConfigurator::fini(){
    int result = SC_FAILED;
    if(SC_STATUS_STARTED == m_status){
        result = SC_NOT_INIT;
    } else if(SC_STATUS_IDLE == m_status) {
        result = SC_NOT_INIT;
    } else if((SC_STATUS_INITIALIZED == m_status) || (SC_STATUS_FAILED == m_status)) {
        m_logger.debug("Service %s - Finalizing Components",m_name.c_str());
    
        ComponentList::reverse_iterator rit;
        for(rit = m_componentList.rbegin(); rit != m_componentList.rend(); rit++){
            ComponentConfiguration * component = (*rit)->component;
            // Finalize The Component
            result = component->fini();
            if(0 != result){
                m_logger.warn("Service %s - Failed To Finalize Component %s",m_name.c_str(), component->getName());
            }
            // Clear Init Params
            ComponentConfiguration::Params& iparams = (*rit)->init_params;
            ComponentConfiguration::Params::iterator pit;
            for(pit = iparams.begin(); pit != iparams.end(); pit++){
                 delete (*pit).second;
            }
            iparams.clear();

            // Clear Config Params
            ComponentConfiguration::Params& cparams = (*rit)->config_params;
            for(pit = cparams.begin(); pit != cparams.end(); pit++){
                 delete (*pit).second;
            }
            cparams.clear();
        }
        
        // TODO: Unload All the dynamically loaded libraries
        
        if(0 == result){
            m_logger.debug("Service %s - Components Finalized",m_name.c_str());
            SysLogDaemon().event(EventNames::ServiceFinalize).success(true).notice();
            result = SC_SUCCESS;
            m_status = SC_STATUS_IDLE;
        } else {
            m_logger.debug("Service %s - Components Finalized With Errors",m_name.c_str());
            SysLogDaemon().event(EventNames::ServiceFinalize).success(false).error();
            result = SC_FAILED;
            m_status = SC_STATUS_FAILED;
        }
    }
    return result;
}


/*
 * registerComponent
 *
 * Register a Component: it will be configured without loading it dynamically
 */
int ServiceConfigurator::registerComponent(ComponentConfiguration * component){

    // Create ComponentItem instance
    ComponentItem * compItem = new ComponentItem(); 
    compItem->comp_handle    = 0;
    compItem->component      = component;
    
    // Insert the the component already exists
    const char * name     = component->getName();
    std::pair<ComponentMap::iterator, bool> result = m_components.insert(ComponentMap::value_type(name,compItem));
    if(false == result.second){
        m_logger.error("Service %s - Component %s Already Registered",m_name.c_str(),name);
        delete compItem;
        return -1;
    } 
    
    // Component will be added to the list after, to preserve the order
    // m_componentList.push_back(compItem);
    
    return 0;
}

/*
 * deregisterComponent
 *
 * Deregister a Static Component
 */
int ServiceConfigurator::deregisterComponent(ComponentConfiguration * component){
    
    // Check if the component already exists
    const char * name     = component->getName();
    ComponentMap::iterator it = m_components.find(name);
    if(it == m_components.end()){
        //Do not log since it may cause problem when the process is finishing
        //m_logger.error("Service %s - Component %s is not Registered",m_name.c_str(),name);
        return -1;
    } 

    // Check if it's the same instance
    if((*it).second->component != component){
        //m_logger.error("Service %s - Not the Same Component %s Instance",m_name.c_str(),name);
        return -1;
    }

    // Erase From the Map
    m_components.erase(name);
    
    // Erase From the list
    ComponentList::iterator lit;
    for(lit = m_componentList.begin(); lit != m_componentList.end(); ++lit){
        if(component == (*lit)->component){
            break;    
        }
    }
    if(lit != m_componentList.end()){
        m_componentList.erase(lit);
    } else {
        //m_logger.warn("Service %s - Component %s Not in the list",m_name.c_str(),name);
    }
    
    return 0;
}

/**
 * Init Logging
 */
int ServiceConfigurator::initLogger(){
    
    // Set Default Layout for errors
    Category& rootLogger = Category::getRoot();
    Appender * app = rootLogger.getAppender();
    if(0 != app){
        PatternLayout * layout = new PatternLayout();
        layout->setConversionPattern("%m%n");
        app->setLayout(layout);
    }
    const std::string& log_configfile = m_locations->logPropertiesFile();
    // Initialize Logging Service
    try{
        PropertyConfigurator::configure(log_configfile.c_str());
    } catch(ConfigureFailure& exc){
        LOG_ERROR_MESSAGE("Logging initialization from file " << log_configfile << " failed: " << exc.what());
        return -1;
    }
    m_logger.debug("Service %s - Logging Initialized",m_name.c_str());

    // Configure SysLog helper object
    SysLogInfo::instance().configure(m_name, "syslog-daemon", LOG_DAEMON, "LOG_DAEMON");

    // Configure syslog appenders 
    SysLogDaemon::EnableSystemLog();
    SysLogAuthpriv::EnableSystemLog();
    
    return 0;
}


 
/**
 * Dynamically load a Component
 * @param name      [IN] component name
 * @param libname   [IN] name of the component library
 * @param deplist   [IN] list of dependencies (names of libraries to open 
 * with dlopen)
 * @return pointer to ComponentItem object if successful, 0 otherwise
 */
ServiceConfigurator::ComponentItem * ServiceConfigurator::loadComponent(
    const char *                    name, 
    const char *                    libname, 
    const std::list<std::string>&   deplist){

    ComponentItem * compItem = new ComponentItem(); 

    // First Load all the dependencies
    std::list<std::string>::const_iterator it;
    for(it = deplist.begin(); it != deplist.end(); it++){
        const char * depname = (*it).c_str();
        void * deplib_handle = dlopen(depname, RTLD_LAZY|RTLD_GLOBAL);
        if(0 == deplib_handle){
            m_logger.error("Service %s - failed loading dependency library %s: %s",m_name.c_str(),depname,dlerror());
            delete compItem;
            return 0;
        }
        compItem->dependencies.push_back(deplib_handle);
    }
    
    // Load Library
    void * lib_handle = dlopen(libname, RTLD_LAZY|RTLD_GLOBAL);
    if(0 == lib_handle){
        m_logger.error("Service %s - failed loading component library %s: %s",m_name.c_str(),libname,dlerror());
        delete compItem;
        return 0;
    }
    
    typedef ComponentConfiguration* create_component_t();
    typedef void destroy_component_t(ComponentConfiguration*);

    // Get Pointer to the Create Function
    create_component_t* create_component_ptr = (create_component_t*) dlsym(lib_handle, CREATE_COMPONENT_API);
    if(0 == create_component_ptr){
        m_logger.error("Service %s - function create_component not found: %s",m_name.c_str(),dlerror());
        delete compItem;
        return 0;
    }
    // Get Pointer to the Destory Function
    destroy_component_t* destroy_component_ptr = (destroy_component_t*) dlsym(lib_handle, DESTROY_COMPONENT_API);
    if(0 == destroy_component_ptr){
        m_logger.error("Service %s - function destroy_component not found: %s",m_name.c_str(),dlerror());
        delete compItem;
        return 0;
    }
    // Get The ComponentConfiguration Instance
    ComponentConfiguration * component =  create_component_ptr();
    
    // Store Component Item
    compItem->comp_handle = lib_handle;
    compItem->component   = component;
   
    std::pair<ComponentMap::iterator, bool> result = m_components.insert(ComponentMap::value_type(name,compItem));
    if(false == result.second){
        m_logger.error("Service %s - component %s already registered",m_name.c_str(),name);
        delete compItem;
        return 0;
    } else {
       // Store Component on the List to preserve order
       m_componentList.push_back(compItem);
    }
    
    return compItem;
}

/**
 * log_error_message
 * log an error message to the default logger and also to the file
 * /tmp/<svc-name>-svc-conf-debug.log in case the logger is not yet 
 * configured.
 * @param msg the message to log
 */
void ServiceConfigurator::log_error_message(const char * msg){
    log_message(m_name,"ERROR",msg);
    // try to log on the default logger
    m_logger.error(msg);
}

/**
 * log_debug_message
 * log a debug message to the default logger and also to the file
 * /tmp/<svc-name>-svc-conf-debug.log in case the logger is not yet 
 * configured.
 * @param msg the message to log
 */
void ServiceConfigurator::log_debug_message(const char * msg){
    log_message(m_name,"DEBUG",msg);
    // try to log on the default logger
    m_logger.debug(msg);
}

/*
 * initProcNameVars
 * 
 * Initialize variables for changing the process name. This method should be
 * called in the main function if the process is interested in forking child
 * processing and updating their names
 */
int ServiceConfigurator::initProcNameVars(int argc,char **argv){
    s_argv    = argv;
    s_nameLen = 0;
    for(int i = 0; i < argc; i++){
        s_nameLen += strlen(s_argv[i]) + 1;
    }
    return s_nameLen;
}

/**
 * Change the process name
 * @param name [IN] the name to assign to the process
 * @return the length of the name assigned
 */
int ServiceConfigurator::setProcName(const char * name){
    int result = 0;
    if((s_argv != 0) && (s_nameLen > 0)){ 
        // Clear the command line
        memset(s_argv[0],0,s_nameLen); 
        // Change The Service Name
        char * dest = strncpy(s_argv[0],name,s_nameLen - 1);
        s_argv[0][s_nameLen - 1] = 0;
        // 
        result = (dest - s_argv[0]) /* / sizeof(char) */;
    }
    return result;
}

/*
 * validateName
 * 
 * check if the instance name is valid
 */
bool ServiceConfigurator::validateName(const char * name){
    
    if((0 == name) || 0 == strlen(name)){
        return false;
    }
    if (strspn(name,LEGAL_FILENAME_CHARS) != strlen(name)) {
        return false;
    }
    return true;
}
