/**
 *  Copyright (c) Members of the EGEE Collaboration. 2004.
 *  See http://www.eu-egee.org/partners/ for details on the copyright
 *  holders.
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 *  ServiceConfigurator XML serialization/deserialization
 *
 *  Authors: Paolo Tedesco <paolo.tedesco@cern.ch>
 *  Version info: $Id: ServiceConfigurator_libxml.cxx,v 1.4 2008/01/24 08:43:44 ptedesco Exp $
 *  Release: $Name:  $
 *
 */
#include "glite/data/config/service/ServiceConfigurator.hxx"
#include "glite/data/config/service/ParamValue.hxx"

#include "xmlutils/xpath.hxx"

#include "log4cpp/Category.hh"
#include "log4cpp/PropertyConfigurator.hh"
#include "log4cpp/PatternLayout.hh"

#include "Locations.hxx"

#include <sstream>
#include <stdexcept>

#include <boost/shared_ptr.hpp>

using namespace log4cpp;
using namespace glite::config;
using namespace glite::config::xmlutils;
using boost::shared_ptr;

/**
 * Load component initialization and configuration parameters
 * @param doc [IN] the xml document
 * @param context [IN] the xpath context
 * @param componentXPath [IN] xpath of the current component node
 * @param phase [IN] init, config
 * @param params [OUT] list of parameters that will be populated with xml
 *  parameters data
 */
void loadConfigParams(
    Doc&                            doc,
    XPathContext&                   context,
    const std::string&              componentXPath,
    const std::string&              phase,
    ComponentConfiguration::Params& params) { //throw std::runtime_error 
            
    std::stringstream paramsXPath;
    paramsXPath << componentXPath << "/" << phase << "/param[@name]";

    XPathObject parameters(context,xml_string(paramsXPath.str().c_str()));
    for (int i = 0;i < parameters.length();++i){
        xmlNodePtr parameterNode = parameters->nodesetval->nodeTab[i];
        Attribute parameterName(parameterNode,xml_string("name"));
        if (!parameterName.isValid()){
            throw XmlConfigurationException("Attribute 'name' for 'parameter' node is missing or empty.");
        }
        xmlNodePtr valueNode = getChildElement(parameterNode,xml_string("value"));
        if (0 == valueNode){
            CONFIGURATION_EXCEPTION("Missing 'value' node in parameters");
        }
        XmlNodeValue value(doc,valueNode);
        params.insert(
            ComponentConfiguration::Params::value_type(
                parameterName.c_str(),
                new ParamValue(
                    parameterName.c_str(),
                    value.c_str())));

    }
}

/**
 * Load Service Components
 */
int ServiceConfigurator::loadComponents(){
 
    int result = -1;
    
    const std::string& config_filename = m_locations->propertiesFile();
    
    // Parse the File
    try{
        // open the document
        XmlInit init;
        Doc doc(config_filename);
        XPathContext context(doc);
        // get the components list
        XPathObject components(context,xml_string("/service/components/component"));
        if (components.isEmpty()){
            CONFIGURATION_EXCEPTION("Document '" << config_filename << "' contains no component nodes.");
        }
        // iterate through components
        for (int i=0;i < components.length();++i){
            xmlNodePtr componentNode = components->nodesetval->nodeTab[i];
            // get component name
            Attribute componentName(componentNode,xml_string("name"));
            if (!componentName.isValid()){
                throw XmlConfigurationException("Attribute 'name' for 'component' node is missing or empty.");
            }
            m_logger.debug("Got component '%s'",componentName.c_str());
            // get the component lib
            xmlNodePtr libNode = getChildElement(componentNode,xml_string("lib"));           

            std::stringstream componentXPath;
            componentXPath << "/service/components/component[" << (i+1) << "]";
            ComponentItem * componentItem = 0;
            if (0 != libNode){
                XmlNodeValue libValue(doc,libNode);
                if (!libValue.isValid()){
                    CONFIGURATION_EXCEPTION("Empty 'lib' node for component '" << componentName.c_str() << "'");
                }
                m_logger.debug("Component '%s' - library is '%s'",componentName.c_str(),libValue.c_str());

                // get the dependencies
                std::stringstream depLibsXPath;
                depLibsXPath << componentXPath.str() << "/dependencies/lib";

                XPathObject depLibs(context,xml_string(depLibsXPath.str().c_str()));
                // iterate through dependencies
                std::list<std::string> dependencies;

                for (int libIndex = 0;libIndex < depLibs.length();++libIndex){
                    // get lib name
                    XmlNodeValue libValue(doc,depLibs->nodesetval->nodeTab[libIndex]);
                    if (!libValue.isValid()){
                        CONFIGURATION_EXCEPTION("Empty or missing 'lib' node in dependencies of component '" << componentName.c_str() << "'");
                    }
                    dependencies.push_back(libValue.c_str());
                }
                m_logger.debug("Component %s - got dependencies",componentName.c_str());
                
                // Load Component
                componentItem = loadComponent(componentName.c_str(),libValue.c_str(),dependencies);
                if(0 == componentItem){
                    throw std::runtime_error("Component Not Found");
                }
                m_logger.debug("Component %s loaded",componentName.c_str());
            } else {
                // static component
                ComponentMap::iterator mit = m_components.find(componentName.c_str());
                if(mit == m_components.end()){
                    m_logger.error("Service %s - static component %s is not registered",m_name.c_str(),componentName.c_str());
                    CONFIGURATION_EXCEPTION("Static component " << componentName.c_str() << " is not registered.");
                }
                componentItem = (*mit).second;
                // Check that the component is static
                if(0 != componentItem->comp_handle){
                    m_logger.error("Service %s - %s is not a static component",m_name.c_str(),componentName.c_str());
                    CONFIGURATION_EXCEPTION(componentName.c_str() << " is not a static component");
                }
                
                // Store Component on the List to preserve order
                m_componentList.push_back(componentItem);
            }
            
            // Read The Initilization Parameters
            ComponentConfiguration::Params& initParams = componentItem->init_params;
            loadConfigParams(doc,context,componentXPath.str(),"init",initParams);

            m_logger.debug("Component %s - Initializations Parameter Loaded",componentName.c_str());
            
            // Read The Configuration
            ComponentConfiguration::Params& configParams = componentItem->config_params;
            loadConfigParams(doc,context,componentXPath.str(),"config",configParams);

            m_logger.debug("Component %s - Configuration Parameters Loaded",componentName.c_str());
        }

        // set result to success
        result = 0;
    } catch(const std::exception& exc) {
        m_logger.error("Service %s - got exception: %s",m_name.c_str(),exc.what());
    } catch(...) {
        m_logger.error("Service %s - Got unknown exception",m_name.c_str());
    }
    return result;
}

/**
 * Load Service Components
 */
int ServiceConfigurator::reloadConfiguration(){
 
    int result = -1;
    
    const std::string& config_filename = m_locations->propertiesFile();
    
    // Parse the File
    try{
        // open the document
        XmlInit init;
        Doc doc(config_filename);
        XPathContext context(doc);
        // get the components list
        XPathObject components(context,xml_string("/service/components/component"));
		if (components.isEmpty()){
            CONFIGURATION_EXCEPTION("Document '" << config_filename << "' contains no component nodes.");
        }
        // iterate through components
        for (int i=0;i < components.length();++i){
            xmlNodePtr componentNode = components->nodesetval->nodeTab[i];
            // get component name
            Attribute componentName(componentNode,xml_string("name"));
            if (!componentName.isValid()){
                throw XmlConfigurationException("Attribute 'name' for 'component' node is missing or empty.");
            }
            m_logger.debug("Service '%s' - reloading configuration for component '%s'",m_name.c_str(),componentName.c_str());

            std::stringstream componentXPath;
            componentXPath << "/service/components/component[" << (i+1) << "]";

            // Get the Component
            ComponentMap::const_iterator it = m_components.find(componentName.c_str());
            if(it == m_components.end()){
                CONFIGURATION_EXCEPTION("Cannot add new components duuring reconfiguration");
            }
            ComponentItem * componentItem = (*it).second;
            if(0 == componentItem){
                CONFIGURATION_EXCEPTION("Component '" << componentName.c_str() << "' not found.");
            }
            // Reload The Configuration
            ComponentConfiguration::Params& configParams = componentItem->config_params;
            configParams.clear();
            loadConfigParams(doc,context,componentXPath.str(),"config",configParams);
        }
        result = 0;
    } catch(const std::exception& exc) {
        m_logger.error("Service %s - Got Exception %s",m_name.c_str(),exc.what());
    } catch(...) {
        m_logger.error("Service %s - Got Unknown Exception",m_name.c_str());
    }
    return result;
}
