/**
 *  Copyright (c) Members of the EGEE Collaboration. 2004-2008.
 *  See http://www.eu-egee.org/partners/ for details on the copyright
 *  holders.
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 *  SysLog.cxx
 *
 *  Authors: Paolo Tedesco <paolo.tedesco@cern.ch>
 *  Version info: $Id: SysLog.cxx,v 1.5 2009/12/03 17:18:01 molnarzs Exp $
 *  Release: $Name:  $
 *
 */

#include <glite/data/config/service/SysLog.hxx>
#include <glite/data/config/service/SysLogInfo.h>
#include "SysLogValidator.hxx"
#include <log4cpp/Category.hh>
#include <time.h>

using namespace glite::config;
using namespace log4cpp;

namespace {
#ifdef _DEBUG
    // identifier for log messages (debugging purposes)
    static unsigned long message_number = 0;    
#endif    


} // namespace

/**-----------------------------------------------------------------------------
 * EventNames
 * collection of descriptive strings of event names to be used in system logs.
------------------------------------------------------------------------------*/

/**
 * service was started
 */
const char * const EventNames::ServiceStart = "org.glite.data.service.start";
/**
 * service was re-configured
 */
const char * const EventNames::ServiceReconfigure = "org.glite.data.service.reconfigure";
/**
 * service was stopped
 */
const char * const EventNames::ServiceStop = "org.glite.data.service.stop";
/**
 * service was finalized
 */
const char * const EventNames::ServiceFinalize = "org.glite.data.service.finalize";

/**-----------------------------------------------------------------------------
 * SysLog
 * Base class for system log helpers
------------------------------------------------------------------------------*/

const std::string SysLog::FileCategoryName = "syslog";

const std::string SysLog::FormatString = "ts=%d{%FT%T%z}%m";

/**
 * Constructor
 * @param categoryName [IN] the name of the log4cpp category
 */
SysLog::SysLog(
    const std::string&  categoryName,
    const std::string&  facility,
    bool                fileLog):
        m_categoryName(categoryName),
        m_facility(facility),
        m_fileLog(fileLog), 
        m_status(0),
        m_statusInit(false){
}

/**
 * Destructor
 */
SysLog::~SysLog(){
}

/**
 * Set the session or request id
 * @param id [IN] the session or request id
 * @return a reference to the object itself
 */
SysLog& SysLog::id(const std::string& id){
    m_id = SysLogValidator::instance().validate_value(id);
    return (*this);
}

/**
 * Set the external request or session id
 * @param guid [IN] the external request or session id
 * @return a reference to the object itself
 */
SysLog& SysLog::guid(const std::string& guid){
    m_guid = SysLogValidator::instance().validate_value(guid);
    return (*this);
}

/**
 * Set the event name
 * @param event [IN] the event name
 * @return a reference to the object itself
 */
SysLog& SysLog::event(const std::string& event){
    SysLogValidator::instance().validate_event_name(event,m_event);
    return (*this); 
}

/**
 * Set the event status to 0 or -1 (generic codes)
 * @param success [IN] success flag
 * @return a reference to the object itself
 */
SysLog& SysLog::success(bool success){
    m_statusInit = true;
    m_status = success ? 0 : -1;
    return (*this); 
}

/**
 * Set the event status to given value
 * @param status [IN] event status value: 
 *  < 0 -> failure
 *  >= 0 -> success
 * @return a reference to the object itself
 */
SysLog& SysLog::status(int status){
    m_statusInit = true;
    m_status = status;
    return (*this); 
}

/**
 * Set the message
 * @param message [IN] the message
 * @return a reference to the object itself
 */
SysLog& SysLog::msg(const std::string& message){
    m_message = SysLogValidator::instance().validate_value(message);
    return (*this); 
}

/**
 * Add a generic name-value pair
 * @param name [IN] name
 * @param value [IN] value
 * @return a reference to the object itself
 */
SysLog& SysLog::add(const std::string& name,const std::string& value){
    m_nameValuePairs[SysLogValidator::instance().validate_name(name)] = 
        SysLogValidator::instance().validate_value(value);
    return (*this); 
}

/**
 * add_common
 * add common parts of log message
 * @param stream [IN] the log stream
 */
void SysLog::add_common(log4cpp::CategoryStream& stream){
    // session id [5]
    if (!m_id.empty()){
        stream << " id=\"" << m_id << "\"";
    }
    // external session id [10.4]
    if (!m_guid.empty()){
        stream << " guid=\"" << m_guid << "\"";
    }
    // event name [7]
    if (!m_event.empty()){
        stream << " event=" << m_event;
    }
    // status [8]
    if (m_statusInit){
        stream << " status=" << m_status;
    }
    // message [9]
    if (!m_message.empty()){
        stream << " msg=\"" << m_message << "\"";
    }
    // other parameters
    for (NameValuePairs::const_iterator it = m_nameValuePairs.begin();it != m_nameValuePairs.end();++it){
        stream << " " << it->first << "=\"" << it->second << "\"";
    }
#ifdef _DEBUG
    stream << " msg_dbg_id=" << message_number++;
#endif    
}

/**
 * Log a message to the system log with given priority
 * @param priority [IN] the log priority
 */
void SysLog::log(log4cpp::Priority::PriorityLevel priority){
    log4cpp::Category& syslog_category(log4cpp::Category::getInstance(m_categoryName));
    if (syslog_category.isPriorityEnabled(priority)){
        log4cpp::CategoryStream syslog_stream = syslog_category.getStream(priority);
        add_common(syslog_stream);
        syslog_stream.flush();
    }
    if (m_fileLog){
        log4cpp::Category& file_category(log4cpp::Category::getInstance(SysLog::FileCategoryName));
        if (file_category.isPriorityEnabled(priority)){
            log4cpp::CategoryStream file_stream = file_category.getStream(priority);
            // tag and facility [4.1]
            file_stream << "prog=" << SysLogInfo::instance().serviceId() << ":";
            file_stream << " facility=" << m_facility;
            add_common(file_stream);
            file_stream.flush();
        }
    }
}
