/*
 * Copyright (c) Members of the EGEE Collaboration. 2004.
 * See http://www.eu-egee.org/partners/ for details on the copyright holders.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *     http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 *  main file for the GLite Data Conf Service ConfigGenerator 
 *
 *  Authors: Paolo Badino <paolo.badino@cern.ch>, Paolo Tedesco <paolo.tedesco@cern.ch>
 *  Version info: $Id: ConfigGenerator.cxx,v 1.9 2009/10/29 10:22:44 szamsu Exp $
 *  Release: $Name:  $
 *  
 */
#include "Service.hxx"
#include "ConsoleInputHandler.hxx"
#include "PropertyFileInputHandler.hxx"
#include "PropertyInputHandler.hxx"

#include <stdlib.h>
#include <cstring>
#include <stdio.h>
#include <unistd.h>
#include <vector>
#include "log4cpp/Category.hh"

using namespace log4cpp;
using namespace glite::config::generator;

namespace {
const char * PROGNAME   = "glite-data-config-generator";
const char * VERSION    = CONFIG_SERVICE_VERSION;
}

/*
 * print_usage
 * 
 * Print Usage
 */
void print_usage(){
    fprintf(stderr,"\n");
    fprintf(stderr,"<%s> Version %s\n",PROGNAME,VERSION);
    fprintf(stderr,"\nUsage:\n");
    fprintf(stderr,"%s -f <config-template> [-o <output_file>] [-l <log_file>] [-p <property_file> | -O <property_name>=<value>][*] [-i] [-L <level>]\n",PROGNAME);
    fprintf(stderr,"where:\n");
    fprintf(stderr,"\t-f config-template: specify the template file to be used to create the\n"\
                   "\t                    configuration\n");
    fprintf(stderr,"\t-o output_file    : the name of the configuration file to create\n");
    fprintf(stderr,"\t-l log_file       : the name of the log configuration file to create\n");
    fprintf(stderr,"\t-p property_file  : use the property file as input. If a mandatory \n"\
                   "\t                    property is missing, the generation fails, unless \n"\
                   "\t                    the interactive mode is enabled. Multiple files \n"\
                   "\t                    can be specified by repeating the -p option\n");
    fprintf(stderr,"\t-O property=value : Set a property from the command line. Multiple \n"
                   "\t                    properties can be specified by repeating the -O \n"
                   "\t                    option and can be mixed with the -p option\n");
    fprintf(stderr,"\t-i                : Run the interacting mode, asking for unset \n"\
                   "\t                    properties, depending on the chosen level\n");
    fprintf(stderr,"\t-L level          : the detail level that should be used. Allowed values\n"\
                   "\t                     are:\n");
    fprintf(stderr,"\t                     low      - only mandatory parameters\n");
	fprintf(stderr,"\t                     normal   - increase the level by adding some \n"\
                   "\t                                parameters that usually you want to set\n");
	fprintf(stderr,"\t                     advanced - increase the level by adding parameters\n"\
                   "\t                                that has a reasonable default\n");
	fprintf(stderr,"\t                     expert   - include all the parameters\n");
	fprintf(stderr,"\t                    Accordingly to the selected level, the user would be\n"
                   "\t                    asked only for unset parameters\n");
    fprintf(stderr,"\t-d                : Enable debug logging\n");
    fprintf(stderr,"\n");
}

/*
 * main
 * 
 * main function
 */
int main(int argc, char * argv[]) {

    char * svc_configfile = 0;
    char * outfile        = 0;
    char * logfile        = 0;
    bool   debugLog       = false;
    bool   interactive    = true;
    int    level          = -1;
    
    typedef std::vector<InputHandler*> InputHandlers;
    InputHandlers handlers;
    
    int c;
    while ((c = getopt (argc, argv, "hdf:o:l:p:O:L:i")) != -1)
    switch (c){
        case 'f':
            svc_configfile = optarg;
        break;
        case 'o':
            outfile = optarg;
        break;
        case 'l':
            logfile = optarg;
        break;
        case 'p':
            handlers.push_back(new PropertyFileInputHandler(optarg));
            interactive = false;
        break;
        case 'O':
            handlers.push_back(new PropertyInputHandler(optarg));
            interactive = false;
        break;
        case 'i':
            interactive = true;
        break;
        case 'L':
            if(0 == strcasecmp(optarg,"low")){
                level = 0;
            } else if(0 == strcasecmp(optarg,"normal")){
                level = 1;
            } else if(0 == strcasecmp(optarg,"advanced")){
                level = 2;
            } else if(0 == strcasecmp(optarg,"high")){
                level = 3;
            } else {
                // Invalid Value
                level = -2;
            }
        break;
        case 'd':
            debugLog = true;
        break;
        case 'h':
        default:
            print_usage();
            return -1;
        break;
    }
    if(0 == svc_configfile){
       print_usage();
       return -1;
    }
    if((level < -1) || (level > 3)){
        print_usage();
        return -1;
    }

    // Set Log Properties
    if(debugLog){
        Category::setRootPriority(Priority::DEBUG);
    } else {
        Category::setRootPriority(Priority::INFO);
    }
    
    Service             svc;
    LogGenerator        log;
    // Set the Console Input Handler
    if(handlers.empty() || (true == interactive)){
        handlers.push_back(new ConsoleInputHandler(level));
    }
    
    // Parse the Configuration Template
    int result = svc.parse(svc_configfile);
    if(0 == result){
        // Get values from Input
        InputHandlers::iterator it;
        for(it = handlers.begin(); it != handlers.end(); ++it){
            result = (*it)->getValues(&svc);
            if(0 != result){
                break;
            }
        }
    }
    if(0 == result){
        // Get Generate the configuration file
        result = svc.generate(outfile);
    }
    if(0 == result){
        
        // Get Log values from Input
        InputHandlers::iterator it;
        for(it = handlers.begin(); it != handlers.end(); ++it){
            result = (*it)->getLogValues(&svc,&log);
            if(0 != result){
                break;
            }
        }
    }
    // FIX
    log.setServiceName(svc.getName());
    if(0 == result){
        // Generate the logging configuration file
        result = log.generate(logfile);
    }
    
    InputHandlers::iterator it;
    for(it = handlers.begin(); it != handlers.end(); ++it){
        if((*it) != 0){
            delete (*it);
        }
    }
    handlers.clear();

    return result;
}
