/*
 * Copyright (c) Members of the EGEE Collaboration. 2004.
 * See http://www.eu-egee.org/partners/ for details on the copyright holders.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *     http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 *  GLite Data Conf Service ConfigGenerator : ConsoleInputHandler.cxx
 *
 *  Authors: Paolo Badino <paolo.badino@cern.ch>, Paolo Tedesco <paolo.tedesco@cern.ch>
 *  Version info: $Id: ConsoleInputHandler.cxx,v 1.11 2009/10/29 10:22:44 szamsu Exp $
 *  Release: $Name:  $
 *  
 */
#include "ConsoleInputHandler.hxx"

#include <stdio.h>
#include <unistd.h>
#include <cstring>
#include <cstdlib>

#include "log4cpp/Category.hh"
using namespace log4cpp;
using namespace glite::config::generator;

/*
 * ConsoleInputHandler
 * 
 * Constructor
 */
ConsoleInputHandler::ConsoleInputHandler(int level):
    _logger(Category::getInstance("glite-config-generator")),
    _level(level),
    _mandatoryOnly(false),
    _advancedMode(false),
    _expertMode(false){
}

/*
 * includeParam
 * 
 * Check if the parameter required configuration
 */
bool ConsoleInputHandler::includeParam(Param * param, bool mandatory_only, bool advanced_mode, bool expert_mode){
    
    if(true == param->getInclude()){
        return false;
    }
    // Check if the parameter should be used
    if(true == param->getFixed()){
        return false;
    }
    if(false == param->getMandatory()){
        if(true == mandatory_only){
            return false;
        }
        switch(param->getLevel()){
            case Param::LEVEL_ADVANCED:
                if(false == advanced_mode){
                    return false;
                }
            break;
            case Param::LEVEL_EXPERT:
                if(false == expert_mode){
                    return false;
                }
            break;
            default:
            break;
        }
    }
    return true;
}
                
/*
 * getValues
 * 
 * Get the Configuration Values for the Service
 */
int ConsoleInputHandler::getValues(Service * service){

    int result = 0;
    char char_rsp[80];
    
    switch(_level){
    case 0:
        _mandatoryOnly = true;
        _advancedMode  = false;
        _expertMode    = false;
    break;
    case 1:
        _mandatoryOnly = false;
        _advancedMode  = false;
        _expertMode    = false;
    break;
    case 2:
        _mandatoryOnly = false;
        _advancedMode  = true;
        _expertMode    = false;
    break;
    case 3:
        _mandatoryOnly = false;
        _advancedMode  = true;
        _expertMode    = true;
    break;
    default:
        fprintf(stdout,"\nPlease chose the level of details you what to use:\n");
        fprintf(stdout,"\t0 - Low     : only mandatory parameters\n");
        fprintf(stdout,"\t1 - Normal  : increase the level by adding some parameters \n"\
                       "\t              that usually you want to set\n");
        fprintf(stdout,"\t2 - Advanced: increase the level by adding parameters that\n"\
                       "\t      has a reasonable default\n");
        fprintf(stdout,"\t3 - Expert  : include all the parameters\n");
        fprintf(stdout,"\nPlease enter your choice: [Default value: 1 (Normal)]");
        fflush(stdout);
        char_rsp[0] = '1';
        fgets(char_rsp, 80, stdin);

        switch(char_rsp[0]){
        case '0':
            _mandatoryOnly = true;
            _advancedMode  = false;
            _expertMode    = false;
        break;
        case '2':
            _mandatoryOnly = false;
            _advancedMode  = true;
            _expertMode    = false;
        break;
        case '3':
            _mandatoryOnly = false;
            _advancedMode  = true;
            _expertMode    = true;
        break;
        case '1':
        default:
            _mandatoryOnly = false;
            _advancedMode  = false;
            _expertMode    = false;
        break;
        } // end switch
    break;
    } // end switch   
    
    bool ask_for_config = false;
    const Service::Components& components = service->getComponents();
    Service::Components::const_iterator it;
    for(it = components.begin(); it != components.end(); it++){
        Component * component = (*it);
        Component::Params& initParams = component->getInitParams();
        Component::Params& configParams = component->getConfigParams();
        Component::Params::const_iterator pit;
            
        // Check if the Component should be configured
        if(initParams.size() > 0){
            for(pit = initParams.begin(); pit != initParams.end(); pit ++){
                Param * param = (*pit);
                if(true == includeParam(param, _mandatoryOnly,_advancedMode ,_expertMode)){
                    ask_for_config = true;
                    break;
                }
            }
        }
        if(ask_for_config == false && configParams.size() > 0){
            for(pit = configParams.begin(); pit != configParams.end(); pit ++){
                Param * param = (*pit);
                if(true == includeParam(param, _mandatoryOnly,_advancedMode ,_expertMode)){
                    ask_for_config = true;
                    break;
                }
            }
        }
        if(ask_for_config == true){
            break;
        }
    }
    if(ask_for_config == true){
        printf("\n********************************************************************************\n");
        printf(  "* Getting Configuration Parameter Values                                       *\n");
        printf(  "********************************************************************************\n\n");
    } else {
        return 0;
    }
    
    for(it = components.begin(); it != components.end(); it++){
        Component * component = (*it);
        Component::Params& initParams = component->getInitParams();
        Component::Params& configParams = component->getConfigParams();
        Component::Params::const_iterator pit;
            
        // Check if the Component should be configured
        bool has_init_params = false;
        bool has_config_params = false;
        // Check Init Params
        if(initParams.size() > 0){
            for(pit = initParams.begin(); pit != initParams.end(); pit ++){
                Param * param = (*pit);
                if(true == includeParam(param, _mandatoryOnly,_advancedMode ,_expertMode)){
                    has_init_params = true;
                    break;
                }
            }
        }
        // Check Config Params
        if(configParams.size() > 0){
            for(pit = configParams.begin(); pit != configParams.end(); pit ++){
                Param * param = (*pit);
                if(true == includeParam(param, _mandatoryOnly,_advancedMode ,_expertMode)){
                    has_config_params = true;
                    break;
                }
            }
        }
        if((true == has_init_params) || (true == has_config_params)){
            printf("\n--------------------------------------------------------------------------------\n");
            printf(  "Configuring Component %s\n",component->getName());
            printf(  "--------------------------------------------------------------------------------\n");
            printf("\nLibrary: %s\n",component->getLibrary());
            printf("\nDescription:\n\n%s\n",component->getDescription());
            
            if(true == has_init_params){
                printf("\n--------------------------------------------------------------------------------\n");
                printf(  "Configuring Initialization Parameters:\n");
                printf(  "--------------------------------------------------------------------------------\n");
                for(pit = initParams.begin(); pit != initParams.end(); pit ++){
                    Param * param = (*pit);
                    if(true == includeParam(param, _mandatoryOnly,_advancedMode ,_expertMode)){
                        setParameterValue(param);
                    }
                }
            }
            if(true == has_config_params){
                printf("\n--------------------------------------------------------------------------------\n");
                printf(  "Configuring Configuration Parameters:\n");
                printf(  "--------------------------------------------------------------------------------\n");
                for(pit = configParams.begin(); pit != configParams.end(); pit ++){
                    Param * param = (*pit);
                    if(true == includeParam(param, _mandatoryOnly,_advancedMode ,_expertMode)){
                        setParameterValue(param);
                    }
                }
            }
        }
    }
     
    return result;
}


/*
 * getLogValues
 * 
 * Get the Logging Configuration Values for the Service
 */
int ConsoleInputHandler::getLogValues(Service * service, LogGenerator * log){
   
    int result = 0;
   
    // FIX
    //log->setServiceName(service->getName());
   
    char str_rsp[1024];
    
    // Check if properties are already set
    if((true == log->isLogFileNameSet()) && 
      ((true == log->isPrioritySet()) || ((false == log->isPrioritySet()) && 
                                          (_mandatoryOnly == true)))){
           return result;
    }
    if((false == log->isPrioritySet()) && (_mandatoryOnly == true)){
        log->setPriority("INFO");
    }
    printf("\n********************************************************************************\n");
    printf(  "* Getting Logging Configuration Values                                       *\n");
    printf(  "********************************************************************************\n\n");
    if(false == log->isPrioritySet()){
        fprintf(stdout,"\nEnter Minimal Log Priority Level. [Default value: 1 (INFO)]\n");
        fprintf(stdout,"\t0 - DEBUG\n");
        fprintf(stdout,"\t1 - INFO\n");
        fprintf(stdout,"\t2 - WARNING\n");
        fprintf(stdout,"\t3 - ERROR\n");
        fprintf(stdout,"\t4 - CRITICAL\n");
        fprintf(stdout,"\t5 - ALERT\n");
        fprintf(stdout,"\t6 - FATAL\n");
        fflush(stdout);
        strcpy(str_rsp,"1");
        fgets(str_rsp, 1024, stdin);
        // Remove Last Line Break
        char * str_rsp_end = strrchr (str_rsp, '\n'); 
        *str_rsp_end = '\0';
        const char * priority = "INFO";
        if(0 != strlen(str_rsp)){
            if(str_rsp[strspn(str_rsp," \t")] != '\n'){
                int val = atoi(str_rsp);
                switch(val){
                    case 0:
                        priority = "DEBUG";
                    break;
                    case 1:
                        priority = "INFO";
                    break;
                    case 2:
                        priority = "WARN";
                    break;
                    case 3:
                        priority = "ERROR";
                    break;
                    case 4:
                        priority = "CRIT";
                    break;
                    case 5:
                        priority = "ALERT";
                    break;
                    case 6:
                        priority = "FATAL";
                    break;
                }
            }
        }
        
        log->setPriority(priority);
    }
    
    if(false == log->isLogFileNameSet()){
        fprintf(stdout,"\nEnter Destination Log File name. [Default value: /tmp/%s.log]",service->getName());
        fflush(stdout);
        sprintf(str_rsp,"/tmp/%s",service->getName());
        fgets(str_rsp, 1024, stdin);
        // Remove Last Line Break
        char * str_rsp_end = strrchr (str_rsp, '\n'); 
        *str_rsp_end = '\0';
        if(0 != strlen(str_rsp)){
            if(str_rsp[strspn(str_rsp," \t")] != '\n'){
                log->setLogFileName(str_rsp);
            }
        }
    }
    if(true == _advancedMode){
        fprintf(stdout,"\nDo you want to enable file Rolling? (Y/N) [N]");
        fflush(stdout);
        fgets(str_rsp, 1024, stdin);
        
        bool do_rolling = ((str_rsp[0] == 'Y') || (str_rsp[0] == 'y'))?true:false;
       
        log->setRollFile(do_rolling);
    }
    return result;
}
    
/*
 * setParameterValue
 * 
 * Ask the user the values of a single component
 */
void ConsoleInputHandler::setParameterValue(Param * param){
    
    char str_rsp[1024];
    
    // Display Parameter Information
    int l = 0;
    int dash_len = strlen(param->getName()) + 16;
    for(l = 0; l < dash_len;l++){
        printf("-");
    }
    printf("\n| Parameter <%s> |\n",param->getName());
    for(l = 0; l < dash_len;l++){
        printf("-");
    }
    printf("\n");
    printf("\nType: %s\n",param->getType());
    printf("\nMandatory: %s\n",(true==param->getMandatory())?"true":"false");
    printf("\nDescription: %s\n",param->getDescription());
    if(false == param->getMandatory()){
        // Ask if the your want to set the parameter
        fprintf(stdout,"\nDo you want to include that parameter? (Y/N) [Y]");
        fflush(stdout);
        fgets(str_rsp, 1024, stdin);
        param->setInclude(((str_rsp[0] == 'N') || (str_rsp[0] == 'n'))?false:true);
    } else {
        param->setInclude(true);
    }
    if(true == param->getInclude()){
        fprintf(stdout,"\nPlease enter the value for parameter : %s ",param->getName());
        if(false == param->getMandatory()){
            fprintf(stdout,"[Default Value: %s]",param->getValue());
        } else {
            fprintf(stdout,": ",param->getValue());
        }
        fflush(stdout);
        fgets(str_rsp, 1024, stdin);
        // Remove Last Line Break
        char * str_rsp_end = strrchr (str_rsp, '\n'); 
        *str_rsp_end = '\0';
        if(0 != strlen(str_rsp)){
            if(str_rsp[strspn(str_rsp," \t")] != '\n'){
                // TODO Validate Input
                param->setValue(str_rsp);
            }
        } 
        // DEBUG Trace
        _logger.debug("Parameter %s set to %s",param->getName(), param->getValue());
    }
}
