/*
 * Copyright (c) Members of the EGEE Collaboration. 2004.
 * See http://www.eu-egee.org/partners/ for details on the copyright holders.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *     http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 *  GLite Data Conf Service ConfigGenerator : PropertyFileInputHandler.cxx
 *
 *  Authors: Paolo Badino <paolo.badino@cern.ch>
 *  Version info: $Id: PropertyFileInputHandler.cxx,v 1.9 2009/10/29 10:22:44 szamsu Exp $
 *  Release: $Name:  $
 *
 *  
 */
#include "PropertyFileInputHandler.hxx"

#include <stdio.h>
#include <unistd.h>
#include <fstream>
#include <cstring>

#include "log4cpp/Category.hh"
using namespace log4cpp;
using namespace glite::config::generator;

/*
 * ConsoleInputHandler
 * 
 * Constructor
 */
PropertyFileInputHandler::PropertyFileInputHandler(const char * file_name):_logger(Category::getInstance("glite-config-generator")){
   _fileName = file_name;
}

/*
 * getValues
 * 
 * Get the Configuration Values for the Service
 */
int PropertyFileInputHandler::getValues(Service * service){

    return parseProperties(service, 0, true);
}


/*
 * getLogValues
 * 
 * Get the Logging Configuration Values for the Service
 */
int PropertyFileInputHandler::getLogValues(Service * service, LogGenerator * log){
    return parseProperties(service, log, false);
}

/*
 * getValues
 * 
 * Get the Configuration Values for the Service
 */
int PropertyFileInputHandler::parseProperties(Service * service, LogGenerator * log, bool serviceProperty){
   
    int result;
    // Read the configuration File
    std::ifstream configFile(_fileName.c_str());
    if(configFile.is_open()){
        char line[2048]; 
        std::string section;
        while(configFile.good()){
            std::string param_name,param_value;
            configFile.getline(line,2048);
            // Look for comments (starting with ; or #)
            size_t pos = strspn(line, " \t");
            if((line[pos] == ';') || (line[pos] == '#')){
                // Skip comment
                continue;
            }
            // Check if it's a section
            if(line[pos] == '['){
                ++pos;
                // A section starts
                char * pch = strchr(line + pos,']');
                if(0 == pch){
                    // Invalid section
                    continue;
                } else {
                    // Get the section name
                    section = std::string(line + pos, pch);
                }
            } else { // it should be a parameter
                // Check Separator
                char * pch = strchr(line,'=');
                if(0 == pch){
                    // Invalid parameter. Skip it
                    continue;
                }
                *pch = '\0';
                // Get Property Name
                char * name = line;
                // Get Property Value
                char * value = pch + 1;
                
                // EatWhites on name
                pos = strspn(name, " \t");
                char * name_begin = name + pos;
                pos = strcspn(name_begin, " \t");
                name_begin[pos] = 0;
 
                // Generate Name based on the section
                if(section.empty()){
                    param_name = name_begin;
                } else {
                    param_name = section + "." + name_begin;
                }
                // EatWhites on value
                pos = strspn(value, " \t");
                char * value_begin = value + pos;

                // Check if line ends with a '\'
                if(value_begin[strlen(value_begin)-1] == '\\'){
                    if(strlen(value_begin) > 1){
                        param_value = std::string(value_begin, strlen(value_begin)-1);
                    }
                    do{
                        param_value += "\n";
                        configFile.getline(line,2048);
                        if(line[strlen(line)-1] == '\\'){
                            // Another Line 
                            if(strlen(line) > 1){
                                param_value += std::string(line, strlen(line)-1);
                            }
                        } else {
                            // Last Line
                            param_value += line; 
                            break;
                        }
                    } while(configFile.good());
                } else {
                    pos = 0;
                    char * str = value_begin;

                    do{
                        // Check Blacks
                        pos = strcspn(str, " \t");
                        if(pos == strlen(str) || 0 == pos){
                            // All blacks
                            str += pos;
                            break;
                        }
                        // Increase pos
                        str += pos;
                        // Check Whites
                        pos = strspn(str, " \t");
                        if(pos == strlen(str) || 0 == pos){
                            // Only whites
                            break;
                        }
                        str += pos;
                    }while(pos != 0);
                                                                                        
                    value_begin[str - value_begin] = 0;
                    // Store value
                    param_value = value_begin;
                }
                
                // Set value for the given the Property
                if(serviceProperty){
                    setProperty(param_name.c_str(), param_value.c_str(), service);
                } else {
                    setLogProperty(param_name.c_str(), param_value.c_str(), service,log);
                }
            } // end if
        }
        result = 0;
    } else {
        _logger.error("Failed to open File %s",_fileName.c_str());
        result = -1;
    }
    
    return result;
}

// Set value for the given the Property
int PropertyFileInputHandler::setProperty(const char * name, const char * value, Service * service){
    
    // Get the Component Name and the Property Name
    const char * dot_separator = strrchr(name, '.'); 
    if(0 == dot_separator){
        // Skip that line
        return 0;
    }
    std::string componentName;
    std::string propertyName;
    componentName.append(name,dot_separator - name);
    propertyName.append(dot_separator + 1);
   
    // DEBUG Trace
    _logger.debug("Found Component Name %s - Property %s",componentName.c_str(), propertyName.c_str());
        
    // Get the Service Components
    const Service::Components& components = service->getComponents();
    Service::Components::const_iterator it;
    for(it = components.begin(); it != components.end(); it++){
        Component * component = (*it);
        if(0 == strcasecmp(componentName.c_str(), component->getName())){
            // DEBUG Trace
            _logger.debug("Found Component %s",componentName.c_str());
                
            Component::Params& initParams = component->getInitParams();
            Component::Params::const_iterator pit;
            for(pit = initParams.begin(); pit != initParams.end(); pit++){
                Param * param = (*pit);
                if(0 == strcasecmp(propertyName.c_str(), param->getName())){
                    if(false == param->getFixed()){
                        param->setValue(value);
                        param->setInclude(true);
    
                        // DEBUG Trace
                        _logger.debug("Setting Component %s Init Property %s to %s",componentName.c_str(), propertyName.c_str(), value);
                    }
                }
            }
            // In theroy the same name could also be in the Configuration, 
            // even if if not recommended
            Component::Params& configParams = component->getConfigParams();
            for(pit = configParams.begin(); pit != configParams.end(); pit++){
                Param * param = (*pit);
                if(0 == strcasecmp(propertyName.c_str(), param->getName())){
                    if(false == param->getFixed()){
                        param->setValue(value);
                        param->setInclude(true);
    
                        // DEBUG Trace
                        _logger.debug("Setting Component %s Config Property %s to %s",componentName.c_str(), propertyName.c_str(), value);
                    }
                }
            }
            break; 
        }
    }
    
    return 0;
}


/**
 * Set value for the given the Property
 */
int PropertyFileInputHandler::setLogProperty(const char * name, const char * value, Service * service, LogGenerator * log){
    
    // Get the Component Name and the Property Name
    const char * dot_separator = strrchr(name, '.'); 
    if(0 == dot_separator){
        // Skip that line
        return 0;
    }
    std::string componentName;
    std::string propertyName;
    componentName.append(name,dot_separator - name);
    propertyName.append(dot_separator + 1);
   
    // DEBUG Trace
    _logger.debug("Found Component Name %s - Property %s",componentName.c_str(), propertyName.c_str());

    // Check if the Component Property starts with Log
    if(0 != strncasecmp(componentName.c_str(),"log",3)){
        // Skip that line
        return 0;
    }
        
    // Set Service Name    
    log->setServiceName(service->getName());
    
    // Check the Parameters
    if(0 == strcasecmp(propertyName.c_str(), "fileName")){
        log->setLogFileName(value);
        // DEBUG Trace
        _logger.debug("Setting Log FileName to %s",value);
    }
    if(0 == strcasecmp(propertyName.c_str(), "Priority")){
        log->setPriority(value);
        // DEBUG Trace
        _logger.debug("Setting Log Priority to %s",value);
    }
    if(0 == strcasecmp(propertyName.c_str(), "rootCategory")){
        const char * separator = strchr(value, ','); 
        if(0 != separator){
            std::string priorityVal;
            priorityVal.append(value,separator - value);
            log->setPriority(priorityVal.c_str());
         
            // DEBUG Trace
            _logger.debug("Setting Log Priority to %s",priorityVal.c_str());
        }
    }
    if(0 == strcasecmp(propertyName.c_str(), "EnableRolling")){
        
        bool do_rolling = false;
   
        if((0 == strcasecmp(value, "true")) || (0 == strcasecmp(value, "yes"))){
            do_rolling = true;
        }
        log->setRollFile(do_rolling);
    
        // DEBUG Trace
        _logger.debug("Setting Log File Rolling to %s",(do_rolling)?"true":"false");
    }
    
    return 0;
}
