/*
 * Copyright (c) Members of the EGEE Collaboration. 2004.
 * See http://www.eu-egee.org/partners/ for details on the copyright holders.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *     http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 *  GLite Data Conf Service ConfigGenerator : PropertyInputHandler.cxx
 *
 *  Authors: Paolo Badino <paolo.badino@cern.ch>, Paolo Tedesco <paolo.tedesco@cern.ch>
 *  Version info: $Id: PropertyInputHandler.cxx,v 1.5 2009/10/29 10:22:44 szamsu Exp $
 *  Release: $Name:  $
 *
 *  
 */
#include "PropertyInputHandler.hxx"

#include <stdio.h>
#include <unistd.h>
#include <fstream>
#include <cstring>
#include <cstdlib>

#include "log4cpp/Category.hh"
using namespace log4cpp;
using namespace glite::config::generator;

/*
 * ConsoleInputHandler
 * 
 * Constructor
 */
PropertyInputHandler::PropertyInputHandler(const char * key_value):_logger(Category::getInstance("glite-config-generator")){
   parseProperty(key_value);
}

/*
 * getValues
 * 
 * Get the Configuration Values for the Service
 */
int PropertyInputHandler::getValues(Service * service){
    return setProperty(service);
}


/*
 * getLogValues
 * 
 * Get the Logging Configuration Values for the Service
 */
int PropertyInputHandler::getLogValues(Service * service, LogGenerator * log){
    return setLogProperty(service, log);
}

/*
 * getValues
 * 
 * Get the Configuration Values for the Service
 */
int PropertyInputHandler::parseProperty(const char * key_value){
   
    char * line = strdup(key_value);
    
    // Check Separator
    char * pch = strchr(line,'=');
    if(0 == pch){
        // Invalid Key Value Pair
        _logger.warn("Invallid Key Value Pair: %s",key_value);
        free(line);
        return -1;
    }
    *pch = '\0';

    // Get Property Name
    char * name = line;
                                                                        
    // Get Property Value
    char * value = pch + 1;
                                    
    // EatWhites on name
    size_t pos = strspn(name, " \t");
    char * name_begin = name + pos;
    pos = strcspn(name_begin, " \t");
    name_begin[pos] = 0;
    
    // Store Key
    _key = name_begin;
                                                                        
    // EatWhites on value
    pos = strspn(value, " \t");
    char * value_begin = value + pos;
                                                                        
    pos = 0;
    char * str = value_begin;
    do{
        // Check Black
        pos = strcspn(str, " \t");
        str += pos;
        if(pos == strlen(str) || 0 == pos){
            break;
        }
        // Check Whites
        pos = strspn(value_begin, " \t");
        str += pos;
        if(pos == strlen(str) || 0 == pos){
            break;
        }
    }while(pos == 0);
                                                                                
    value_begin[str - value_begin] = 0;
    
    // Store Value
    _value = value_begin;
    
    // Delete Temporary Copy
    free(line);

    return 0;
}

// Set value for the given the Property
int PropertyInputHandler::setProperty(Service * service){
    
    // Get the Component Name and the Property Name
    std::string::size_type dot_pos = _key.find('.');
    if(dot_pos == std::string::npos){
        // Invalid property
        _logger.warn("Unknown property: %s",_key.c_str());
        return 0;
    }
    std::string componentName;
    std::string propertyName;
    componentName.append(_key.substr(0,dot_pos));
    propertyName.append(_key.substr(dot_pos + 1));
   
    // DEBUG Trace
    _logger.debug("Found component name %s - property %s",componentName.c_str(), propertyName.c_str());
        
    // Get the Service Components
    const Service::Components& components = service->getComponents();
    Service::Components::const_iterator it;
    for(it = components.begin(); it != components.end(); it++){
        Component * component = (*it);
        if(0 == strcasecmp(componentName.c_str(), component->getName())){
            // DEBUG Trace
            _logger.debug("Found component %s",componentName.c_str());
                
            Component::Params& initParams = component->getInitParams();
            Component::Params::const_iterator pit;
            for(pit = initParams.begin(); pit != initParams.end(); pit++){
                Param * param = (*pit);
                if(0 == strcasecmp(propertyName.c_str(), param->getName())){
                    if(false == param->getFixed()){
                        param->setValue(_value.c_str());
                        param->setInclude(true);
    
                        // DEBUG Trace
                        _logger.debug("Setting component %s init property %s to %s",componentName.c_str(), propertyName.c_str(), _value.c_str());
                    }
                }
            }
            // In theory the same name could also be in the Configuration, 
            // even if if not recommended
            Component::Params& configParams = component->getConfigParams();
            for(pit = configParams.begin(); pit != configParams.end(); pit++){
                Param * param = (*pit);
                if(0 == strcasecmp(propertyName.c_str(), param->getName())){
                    if(false == param->getFixed()){
                        param->setValue(_value.c_str());
                        param->setInclude(true);
    
                        // DEBUG Trace
                        _logger.debug("Setting component %s config property %s to %s",componentName.c_str(), propertyName.c_str(), _value.c_str());
                    }
                }
            }
            break; 
        }
    }
    
    return 0;
}


/**
 * Set value for the given the Property
 */
int PropertyInputHandler::setLogProperty(Service * service, LogGenerator * log){
    
    // Get the Component Name and the Property Name
    std::string::size_type dot_pos = _key.find('.');
    if(dot_pos == std::string::npos){
        // Invalid property
        _logger.warn("Unknonw log property: %s",_key.c_str());
        return 0;
    }
    std::string componentName;
    std::string propertyName;
    componentName.append(_key.substr(0,dot_pos));
    propertyName.append(_key.substr(dot_pos + 1));
   
    // DEBUG Trace
    _logger.debug("Found component name %s - property %s",componentName.c_str(), propertyName.c_str());

    // Check if the Component Property starts with Log
    if(0 != strncasecmp(componentName.c_str(),"log",3)){
        // Skip that line
        return 0;
    }
        
    // Set Service Name    
    // FIX
    // log->setServiceName(service->getName());
    
    // Check the Parameters
    if(0 == strcasecmp(propertyName.c_str(), "fileName")){
        log->setLogFileName(_value.c_str());
        // DEBUG Trace
        _logger.debug("Setting log fileName to %s",_value.c_str());
    }
    if(0 == strcasecmp(propertyName.c_str(), "priority")){
        log->setPriority(_value.c_str());
        // DEBUG Trace
        _logger.debug("Setting log priority to %s",_value.c_str());
    }
    if(0 == strcasecmp(propertyName.c_str(), "rootCategory")){
        std::string::size_type separator_pos = _value.find(','); 
        if(separator_pos != std::string::npos){
            std::string priorityVal;
            priorityVal.append(_value.substr(0,separator_pos));
            log->setPriority(priorityVal.c_str());
         
            // DEBUG Trace
            _logger.debug("Setting log priority to %s",priorityVal.c_str());
        }
    }
    if(0 == strcasecmp(propertyName.c_str(), "EnableRolling")){
        
        bool do_rolling = false;
   
        if((0 == strcasecmp(_value.c_str(), "true")) || (0 == strcasecmp(_value.c_str(), "yes"))){
            do_rolling = true;
        }
        log->setRollFile(do_rolling);
    
        // DEBUG Trace
        _logger.debug("Setting log file rolling to %s",(do_rolling)?"true":"false");
    }
    
    return 0;
}
