/*
 * Copyright (c) Members of the EGEE Collaboration. 2004.
 * See http://www.eu-egee.org/partners/ for details on the copyright holders.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *     http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 *  GLite Data Conf Service ConfigGenerator : Service.cxx
 *
 *  Authors: Paolo Badino <paolo.badino@cern.ch>, Paolo Tedesco <paolo.tedesco@cern.ch>
 *  Version info: $Id: Service.cxx,v 1.10 2009/10/29 10:22:44 szamsu Exp $
 *  Release: $Name:  $
 */
#include "Service.hxx"
#include <iostream>
#include <cstring>
#include <fstream>
#include "log4cpp/Category.hh"

#include "writer.hxx"
#include "xpath.hxx"

using namespace log4cpp;
using namespace glite::config::generator;
using namespace glite::config::xmlutils;

/*
 * Service
 * 
 * Constructor
 */
Service::Service():_logger(Category::getInstance("glite-config-generator")){
}

/*
 * ~Service
 * 
 * Destructor
 */
Service::~Service(){

    Components::const_iterator it;
    for(it = _components.begin(); it != _components.end(); it++){
        delete (*it);
    }
}

/*
 * parse
 * 
 * Parse the Configuration Template for the given service
 */
int Service::parse(const char * svc_configfile){
     return loadComponents(svc_configfile);
}

/*
 * generate
 * 
 * Generate the Configuration for the given service
 */
int Service::generate(const char * out_name){
     std::string filename;
     if(0 == out_name){
         filename = _name + ".properties.xml";
     } else {
        filename = out_name;
     }
     return writeComponents(filename);
}

/**
 * Load component initialization and configuration parameters
 * @param doc [IN] the xml document
 * @param context [IN] the xpath context
 * @param componentXPath [IN] xpath of the current component node
 * @param phase [IN] init, config
 * @param params [OUT] list of parameters that will be populated with xml
 *  parameters data
 */
void Service::loadParams(
    Doc&                doc,
    XPathContext&       context,
    const std::string&  componentXPath,
    const std::string&  phase,
    Component::Params&  params) { //throw std::runtime_error 
            
    std::stringstream paramsXPath;
    paramsXPath << componentXPath << "/config-template/" << phase << "/param[@name]";

    XPathObject parameters(context,xml_string(paramsXPath.str().c_str()));
    for (int i = 0;i < parameters.length();++i){
        xmlNodePtr paramNode = parameters->nodesetval->nodeTab[i];
        Attribute paramName(paramNode,xml_string("name"));
        if (!paramName.isValid()){
            throw XmlConfigurationException("Attribute 'name' for 'parameter' node is missing or empty.");
        }
        _logger.debug("Got parameter %s",paramName.c_str());
        // Create Param Object
        Param * param = new Param(paramName.c_str());
        // Store Param
        params.push_back(param);

        Attribute type(paramNode,xml_string("type"));
        if (type.isValid()){
            param->setType(type.c_str());
            _logger.debug("Parameter %s type: %s",paramName.c_str(),type.c_str());
        }
        
        Attribute mandatory(paramNode,xml_string("mandatory"));
        if (mandatory.isValid()){
            if(0 == strcasecmp(mandatory.c_str(),"true")){
                param->setMandatory(true);
            } else if(0 == strcasecmp(mandatory.c_str(),"fixed")){
                param->setFixed(true);
                param->setInclude(true);
            } else {
                param->setMandatory(false);
            }
        } else {
            param->setMandatory(false);
        }
        _logger.debug("Parameter %s mandatory flag set",paramName.c_str());

        // Get Advanced Flag
        Attribute advanced(paramNode,xml_string("advanced"));
        if (advanced.isValid()){
            std::string advancedValue = advanced.c_str();
            if(0 == strcasecmp(advancedValue.c_str(),"true")){
                param->setLevel(Param::LEVEL_ADVANCED);
            } else if(0 == strcasecmp(advancedValue.c_str(),"expert")){
                param->setLevel(Param::LEVEL_EXPERT);
            } else {
                param->setLevel(Param::LEVEL_NORMAL);
            }

            // DEBUG Trace
            _logger.debug("Parameter %s advanced flag set",paramName.c_str());

        } else {
            // get level flag
            Attribute level(paramNode,xml_string("level"));
            if (level.isValid()){
                std::string advancedValue = level.c_str();
                if(0 == strcasecmp(advancedValue.c_str(),"advanced")){
                    param->setLevel(Param::LEVEL_ADVANCED);
                } else if(0 == strcasecmp(advancedValue.c_str(),"expert")){
                    param->setLevel(Param::LEVEL_EXPERT);
                } else {
                    param->setLevel(Param::LEVEL_NORMAL);
                }
            }
        }
        xmlNodePtr descriptionNode = getChildElement(paramNode,xml_string("description"));
        if (0 != descriptionNode){
            _logger.debug("Getting parameter %s description.",paramName.c_str());
            XmlNodeValue descriptionValue(doc,descriptionNode);
            param->setDescription(descriptionValue.c_str());
            _logger.debug("Parameter %s description: %s",paramName.c_str(),descriptionValue.c_str());
        }
        xmlNodePtr valueNode = getChildElement(paramNode,xml_string("value"));
        if (0 != valueNode){
            _logger.debug("Getting parameter %s value.",paramName.c_str());
            XmlNodeValue value(doc,valueNode);
            param->setValue(value.c_str());
            _logger.debug("Parameter %s default value: %s",paramName.c_str(),value.c_str());
        }
    }
}

/*
 * loadComponent
 * 
 * Load Service Components
 */
Component * Service::loadComponent(
    Doc&                doc,
    XPathContext&       context,
    const std::string&  componentXPath,
    const std::string&  componentName){
 
    Component * component = new Component(componentName.c_str());
    
    // get library
    std::stringstream libXPath;
    libXPath << componentXPath << "/config-template/lib";
    XPathObject libs(context,xml_string(libXPath.str().c_str()));
    if (0 == libs.length()){
        _logger.debug("Component %s is static",componentName.c_str());
    } else {
        // get lib name
        XmlNodeValue libName(doc,libs->nodesetval->nodeTab[0]);
        component->setLibrary(libName.c_str());
        _logger.debug("Library for Component %s: %s",componentName.c_str(),libName.c_str());
        // get dependencies
        std::stringstream depsXPath;
        depsXPath << componentXPath << "/dependencies/lib";
        XPathObject depLibs(context,xml_string(depsXPath.str().c_str()));
        std::list<std::string> dependencies;
        for (int libIndex = 0;libIndex < depLibs.length();++libIndex){
            // get lib name
            XmlNodeValue libValue(doc,depLibs->nodesetval->nodeTab[libIndex]);
            if (!libValue.isValid()){
                CONFIGURATION_EXCEPTION("Empty or missing 'lib' node in dependencies of component '" << componentName << "'");
            }
            dependencies.push_back(libValue.c_str());
        }
        component->setDependencies(dependencies);
    }
    // get description
    std::stringstream descriptionXPath;
    descriptionXPath << componentXPath << "/description";
    XPathObject descriptions(context,xml_string(descriptionXPath.str().c_str()));
    if (0 != descriptions.length()){
        XmlNodeValue descriptionValue(doc,descriptions->nodesetval->nodeTab[0]);
        component->setDescription(descriptionValue.c_str());
    }

    // DEBUG Trace
    _logger.debug("Getting parameters for component %s",componentName.c_str());
    
    // Get Init Parameters
    Component::Params& initParams = component->getInitParams();
    loadParams(doc,context,componentXPath,"init",initParams);
    
    // DEBUG Trace
    _logger.debug("Component %s init parameters loaded",componentName.c_str());
    
    // Get Config Parameters
    Component::Params& configParams = component->getConfigParams();
    loadParams(doc,context,componentXPath,"config",configParams);

    // DEBUG Trace
    _logger.debug("Component %s config parameters loaded",componentName.c_str());
    
    return component;
}


/**
 * loadComponents
 * 
 * Load Service Components
 */
int Service::loadComponents(const char * svc_configfile){
 
    int result = -1;
    
    // Parse the File
    try{
        // open the document
        XmlInit init;
        Doc doc(svc_configfile);
        XPathContext context(doc);
        // get the service name
        XPathObject services(context,xml_string("/service[@name]"));
        if (1 != services.length()){
            CONFIGURATION_EXCEPTION("Document '" << svc_configfile << "' contains an invalid root node.");
        }
        xmlNodePtr service=services->nodesetval->nodeTab[0];
        Attribute serviceName(service,xml_string("name"));
        if (!serviceName.isValid()){
            throw XmlConfigurationException("No service name specified.");
        }
        _name = serviceName.c_str();
        // get the components list
        XPathObject components(context,xml_string("/service/components/component"));
		if (components.isEmpty()){
            CONFIGURATION_EXCEPTION("Document '" << svc_configfile << "' contains no component nodes.");
        }
        // iterate through components
        for (int i=0;i < components.length();++i){
            xmlNodePtr componentNode = components->nodesetval->nodeTab[i];
            // get component name
            Attribute componentName(componentNode,xml_string("name"));
            if (!componentName.isValid()){
                throw XmlConfigurationException("Attribute 'name' for 'component' node is missing or empty.");
            }
            _logger.debug("Got component '%s'",componentName.c_str());

            std::stringstream componentXPath;
            componentXPath << "/service/components/component[" << (i+1) << "]";
            
            Component * component = loadComponent(doc,context,componentXPath.str(),componentName.c_str()); 
            _components.push_back(component); 
        }

        // set result to success
        result = 0;
    } catch(const std::exception& exc) {
        _logger.error("Service %s - got exception: %s",_name.c_str(),exc.what());
    } catch(...) {
        _logger.error("Service %s - Got unknown exception",_name.c_str());
    }
    return result;
}
    
/*
 * writeComponents
 * 
 * Write Components to output
 */
int Service::writeComponents(const std::string outputFilePath){

    // Check that all the mandatory paramters has been set
      
    // Get the Service Components
    const Service::Components& components = getComponents();
    Service::Components::const_iterator it;
    for(it = components.begin(); it != components.end(); it++){
        Component * component = (*it);
        Component::Params& initParams = component->getInitParams();
        Component::Params::const_iterator pit;
        for(pit = initParams.begin(); pit != initParams.end(); pit++){
            Param * param = (*pit);
            if((true == param->getMandatory()) && (false == param->getInclude())){
                _logger.error("Mandatory Init Parameter %s for Component %s Not set",param->getName(), component->getName());
                return -1;
            }
        }
        Component::Params& configParams = component->getConfigParams();
        for(pit = configParams.begin(); pit != configParams.end(); pit++){
            Param * param = (*pit);
            if((true == param->getMandatory()) && (false == param->getInclude())){
                _logger.error("Mandatory Config Parameter %s for Component %s Not set",param->getName(), component->getName());
                return -1;
            }
        }
    }
    
    int result = 0;
    // Parse the File
    try{
        XmlInit init;
        XmlFileWriter writer(outputFilePath);
        writer.startDocument();
        writer.startElement("service");
        writer.startElement("components");
        
        // Scan Components
        Components::const_iterator it;
        for(it = _components.begin(); it != _components.end(); it++){
            Component * component = (*it);
            
            writer.startElement("component");
            writer.writeAttribute("name",component->getName());
            if(strlen(component->getLibrary()) > 0){
                writer.writeElement("lib",component->getLibrary());
            }
            
            const Component::StringList& dependencies = component->getDependencies();
            if(dependencies.size() > 0){
                writer.startElement("dependencies");
                Component::StringList::const_iterator dit;
                for(dit = dependencies.begin(); dit != dependencies.end(); dit++){
                    writer.writeElement("lib",(*dit).c_str());
                }
                writer.endElement(); // dependencies
            }
            
            writeParams(writer,"init",component->getInitParams());
            writeParams(writer,"config",component->getConfigParams());
            
            writer.endElement(); // component
        } 
        
        writer.endElement(); // components
        writer.endElement(); // service
        writer.endDocument();
    } catch(const std::exception& exc) {
        _logger.error("Service %s - Got Exception %s",_name.c_str(),exc.what());
        result = -1;
    } catch(...) {
        _logger.error("Service %s - Got Unknown Exception",_name.c_str());
        result = -1;
    }
    
    return result;       
}



/*
 * writeParams
 * 
 * Write Params to output
 */
int Service::writeParams(
    XmlFileWriter&              writer,
    const std::string&          phase,
    const Component::Params&    params){
    
    // check that there is at least one parameter that needs to be written
    Component::Params::const_iterator pit;
    if(params.size() > 0){
        for(pit = params.begin(); pit != params.end(); pit ++){
            if((true == (*pit)->getInclude()) ||
                ((false == (*pit)->getInclude()) && (Param::LEVEL_EXPERT != (*pit)->getLevel()))){
                break;
            }
        }
        if(pit == params.end()){
            return 0;
        }
    }

    // create the init/config node
    writer.startElement(phase.c_str()); // phase (init/config)

    // write all the set parameters
    for(pit = params.begin(); pit != params.end(); pit ++){
        Param * param = (*pit);
        if(true == param->getInclude()){
            // Set Param Attributes
            writer.startElement("param");
            writer.writeAttribute("name",param->getName());
            writer.writeElement("value",param->getValue());
            writer.endElement(); // param
        }
    }
    
    // Print a comment for some of the unset parameters
    for(pit = params.begin(); pit != params.end(); pit ++){
        Param * param = (*pit);
        if((false == param->getInclude()) && 
            (Param::LEVEL_EXPERT != param->getLevel())){
            writer.startComment();
            writer.startElement("param");
            writer.writeAttribute("name",param->getName());
            writer.writeElement("value",param->getValue());
            writer.endElement(); // param
            writer.endComment();
        }
    }
    
    writer.endElement(); // phase (init/config)
    
    return 0;
    
}
