/*
 * Copyright (c) Members of the EGEE Collaboration. 2004.
 * See http://www.eu-egee.org/partners/ for details on the copyright holders.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *     http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * glite-config.cxx
 *
 * Authors: Paolo Badino <paolo.badino@cern.ch>
 * Version info: $Id: glite-config.cxx,v 1.8 2008/06/12 14:23:53 ptedesco Exp $
 * Release: $Name:  $
 *
 */
#include "glite/data/config/service/glite-config.h"
#include "glite/data/config/service/glite-config-register.hxx"
#include "glite/data/config/service/ServiceConfigurator.hxx"

#include <pthread.h>
#include "Locations.hxx"
#include "log-message.hxx"

using namespace glite::config;

namespace {
pthread_mutex_t  g_lock; 
int init1 = pthread_mutex_init(&g_lock,0);
}

/*
 * glite_config_initialize
 *
 * Initialize the Service using Glite Configuration
 */
int glite_config_initialize(const char * name) {
    int result = ServiceConfigurator::SC_FAILED;
    
    // Acquire Lock
    int check_locked = pthread_mutex_trylock(&g_lock);
      if(check_locked != 0)
          log_message("trylock","DEBUG","Cannot lock mutex: glite_config_initialize");
    
    try{
        ServiceConfigurator * instance = ServiceConfigurator::instance();
        Locations locations;
        locations.init_locations(name);
        // Initialize the service
        result = instance->init(&locations);
        if(ServiceConfigurator::SC_SUCCESS == result || 
           ServiceConfigurator::SC_ALREADY_INIT == result) {
            // Start the service
            result = instance->start();
        } else if (ServiceConfigurator::SC_ALREADY_STARTED == result) {
            // Already Initailized
            result = 1;   
        }
        if((result < 0) && (0 != instance)){
            instance->fini();
        }
    } catch (const ServiceConfigurationException& exc) {
        // logging is possibly not initialized
        log_message(name,"ERROR",exc.what());
        result = -1;
    } catch(...){
        // Release Lock
        pthread_mutex_unlock(&g_lock);
        throw;
    }
    
    // Release Lock
    pthread_mutex_unlock(&g_lock);
    
    return result;
}

/*
 * glite_config_reconfigure
 *
 * Reconfigure the Service using Glite Configuration
 */
int glite_config_reconfigure(){
    int result = ServiceConfigurator::SC_FAILED;
    
    // Acquire Lock
    int check_locked = pthread_mutex_trylock(&g_lock);
      if(check_locked != 0)
                log_message("trylock","DEBUG","Cannot lock mutex: glite_config_reconfigure");

    
    try {
        // Reconfigure the service
        result = ServiceConfigurator::instance()->reconfig();
    } catch(...){
        // Release Lock
        pthread_mutex_unlock(&g_lock);
        throw;
    }
    
    // Release Lock
    pthread_mutex_unlock(&g_lock);
    
    return result;
}

/*
 * glite_config_finalize
 *
 * Finalize the Service using Glite Configuration
 */
int glite_config_finalize(){
    int result = ServiceConfigurator::SC_FAILED;
    
    // Acquire Lock
    int check_locked = pthread_mutex_trylock(&g_lock);
      if(check_locked != 0)
	        log_message("trylock","DEBUG","Cannot lock mutex: glite_config_finalize");
    
    try {
        ServiceConfigurator * instance = ServiceConfigurator::instance();
            
        // Stop the service
        result = instance->stop();
        
        // Finalize the service
        result = instance->fini();
        
    } catch(...){
        // Release Lock
        pthread_mutex_unlock(&g_lock);
        throw;
    }
    
    // Release Lock
    pthread_mutex_unlock(&g_lock);
    
    return result;
}

/*
 * glite_config_register
 *
 * Register a Static Component into the Glite Service Configuration
 */
int glite_config_register(glite::config::ComponentConfiguration * component){
    int result = ServiceConfigurator::SC_FAILED;
    
    // Acquire Lock
    int check_locked = pthread_mutex_trylock(&g_lock);
      if(check_locked != 0)
          log_message("trylock","DEBUG","Cannot lock mutex: glite_config_register");      

    
    try {
        ServiceConfigurator * instance = ServiceConfigurator::instance();
        
        // Register the Component
        result = instance->registerComponent(component);
        
    } catch(...){
        // Release Lock
        pthread_mutex_unlock(&g_lock);
        throw;
    }
    
    // Release Lock
    pthread_mutex_unlock(&g_lock);
    
    return result;
}
    
/*
 * glite_config_deregiter
 *
 * Deregister a previously registered Static Component
 */
int glite_config_deregister(glite::config::ComponentConfiguration * component){
    int result = ServiceConfigurator::SC_FAILED;
    
    // Acquire Lock
    int check_locked = pthread_mutex_trylock(&g_lock);
      if(check_locked != 0)
          log_message("trylock","DEBUG","Cannot lock mutex: glite_config_deregister");      

    
    try {
        ServiceConfigurator * instance = ServiceConfigurator::instance();
        
        // Deregister the Component
        result = instance->deregisterComponent(component);
    } catch(...){
        // Release Lock
        pthread_mutex_unlock(&g_lock);
        throw;
    }
    
    // Release Lock
    pthread_mutex_unlock(&g_lock);
    
    return result;
}
    
/*
 * glite_proc_name_init
 *
 * Initialize variables for changing the process name. This method should be
 * called in the main function passing the argc and argv arguments. 
 */  
int glite_proc_name_init(int argc, char ** argv){
    return ServiceConfigurator::initProcNameVars(argc,argv);
}

/*
 * glite_set_proc_name
 *
 * Change the process name
 */
int glite_set_proc_name(const char * name){
    return ServiceConfigurator::setProcName(name);
}
