/**
 *  Copyright (c) Members of the EGEE Collaboration. 2004.
 *  See http://www.eu-egee.org/partners/ for details on the copyright
 *  holders.
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 *  GLite Data Config Service libxml helpers
 *
 *  Authors: Paolo Tedesco <paolo.tedesco@cern.ch>
 *  Version info: $Id: base-utils.hxx,v 1.3 2009/12/03 17:18:01 molnarzs Exp $
 *  Release: $Name:  $
 *
 */
 
#ifndef GLITE_DATA_CONFIG_XMLUTILS_BASE_UTILS_HXX_
#define GLITE_DATA_CONFIG_XMLUTILS_BASE_UTILS_HXX_

#define CONFIGURATION_EXCEPTION(msg) {std::stringstream errstr;errstr << msg;throw XmlConfigurationException(errstr.str());}

#include <string>
#include <stdexcept>
#include <sstream>
#include <libxml/parser.h>

namespace glite{
namespace config{
namespace xmlutils{

/**
 * Get the first child element with given name
 */
inline xmlNodePtr getChildElement(xmlNodePtr parent,const xmlChar* name){
    for (xmlNodePtr child = parent->xmlChildrenNode;0 != child;child = child->next){
        if ((XML_ELEMENT_NODE == child->type) && !xmlStrcmp(child->name,name)){
            return child;
        }
    }
    return 0;
}

/**
 * convert a char* to an xmlChar*
 */
inline const xmlChar* xml_string(const char* string){
    return (xmlChar*)string;
}

/**
 * Configuration exception
 */
class XmlConfigurationException : public std::runtime_error{
public:
    /**
     * Constructor
     */
    explicit XmlConfigurationException(
        const std::string& reason):std::runtime_error(reason){}
    /**
     * Destructor
     */
    virtual ~XmlConfigurationException() throw(){}
private:
    // Default constructor not implemented
    XmlConfigurationException();
};


/**
 * Utility for initialization and finalization of xml library
 */
class XmlInit{
public:
    /**
     * Constructor. Initialize the library.
     */
    XmlInit(){
        xmlInitParser();
        LIBXML_TEST_VERSION
    }
    /**
     * Destructor. Finalize the library.
     */
    ~XmlInit(){
        xmlCleanupParser();
    }
protected:
};

/**
 * Base RAII wrapper for libxml2 objects
 * @param XmlType the wrapped pointer
 * @param CleanupFunctionArgument the argument to the cleanup function 
 *  to be called
 */
template <typename XmlType,typename CleanupFunctionArgument = XmlType>
class XmlWrapper{
public:
    /**
     * Destructor.
     */
    virtual ~XmlWrapper(){
        // Invoke cleanup function
        m_cleanupFunction(m_ptr);
    }
    /**
     * -> operator.
     * @return the wrapped pointer
     */
    XmlType* operator -> () {
        return m_ptr;
    }
    /**
     * @return the wrapped pointer
     */
    XmlType* get(){
        return m_ptr;
    }
protected:
    
    // Cleanup function pointer type definition
    typedef void CleanupFunction(CleanupFunctionArgument*);
    
    /**
     * Constructor
     * @param ptr [IN] the pointer to wrap
     * @param function [IN] the cleanup function to call when the object
     *  goes out of scope
     */
    XmlWrapper(XmlType* ptr,CleanupFunction function):m_ptr(ptr),m_cleanupFunction(function){
    }
    
    // pointer
    XmlType* m_ptr;
    // cleanup function
    CleanupFunction * m_cleanupFunction;
};

/**
 * Xml Document
 * Create an Xml document from file.
 */
class Doc : public XmlWrapper<xmlDoc>{
public:
    /**
     * Constructor.
     * Create an empty document.
     */
    Doc() : XmlWrapper<xmlDoc>(xmlNewDoc(xml_string("1.0")),xmlFreeDoc){
    }
    /**
     * Constructor.
     * @param path [IN] path to the document
     */
    Doc(const std::string& path) : XmlWrapper<xmlDoc>(xmlParseFile(path.c_str()),xmlFreeDoc){
        if (0 == m_ptr){
            CONFIGURATION_EXCEPTION("Parsing document '" << path << "' failed.");
        }
    }
    /**
     * Destructor
     */
    virtual ~Doc(){}
};

/**
 * XmlString
 * A wrapper for XmlStrings that explicitly need to be freed.
 */
class XmlString : public XmlWrapper<xmlChar,void>{
public:
    /**
     * Constructor
     * @param value [IN] the xml string
     */
    XmlString(xmlChar* value) : XmlWrapper<xmlChar,void>(value,xmlFree){
    }
    /**
     * Destructor
     */
    virtual ~XmlString(){}
    /**
     * @return true if the string is non null and has non zero length.
     */
    bool isValid(){
        return ((0 != m_ptr) && (0 != xmlStrlen(m_ptr)));
    }
    /**
     * @return char* (for logging)
     */
    const char * c_str(){
        if (0 == m_ptr){
            return "";
        }
        return (char*)m_ptr;
    }
private:
    XmlString();
};

/**
 * Attribute
 * Stores the value of a node attribute
 */
class Attribute : public XmlString{
public:
    /**
     * Constructor
     * @param node [IN] the Xml element
     * @param name [IN] the name of the attribute
     */
    Attribute(xmlNodePtr node,const xmlChar* name) : XmlString(xmlGetProp(node,name)){
    }
    /**
     * Destructor
     */
    virtual ~Attribute(){}
    /**
     * isDefined
     */
    bool isDefined(){
        return (0 != m_ptr);
    }
private:
    Attribute();
};

/**
 * XmlNodeValue
 * Stores the value of a node (internal text node of an element node)
 */
class XmlNodeValue : public XmlString{
public:
    /**
     * constructor
     * @param doc [IN] the xml document
     * @param node [IN] the xml node
     */
    XmlNodeValue(
        Doc&   		doc,
        xmlNodePtr  node) : XmlString(xmlNodeListGetString(doc.get(), node->xmlChildrenNode, 1)){
    }
    /**
     * Destructor
     */
    virtual ~XmlNodeValue(){}
private:
    XmlNodeValue();
};

} // namespace xmlutils
} // namespace config
} // namespace glite

#endif // GLITE_DATA_CONFIG_XMLUTILS_BASE_UTILS_HXX_
