/**
 *  Copyright (c) Members of the EGEE Collaboration. 2004.
 *  See http://www.eu-egee.org/partners/ for details on the copyright
 *  holders.
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 *  GLite Data Config Service libxml helpers
 *
 *  Authors: Paolo Tedesco <paolo.tedesco@cern.ch>
 *  Version info: $Id: writer.hxx,v 1.2 2008/01/24 08:43:44 ptedesco Exp $
 *  Release: $Name:  $
 *
 */

#ifndef GLITE_DATA_CONFIG_XMLUTILS_WRITER_HXX_
#define GLITE_DATA_CONFIG_XMLUTILS_WRITER_HXX_

#include "base-utils.hxx"
#include <libxml/encoding.h>
#include <libxml/xmlwriter.h>

#   if defined(LIBXML_WRITER_ENABLED) && defined(LIBXML_OUTPUT_ENABLED)

namespace glite{
namespace config{
namespace xmlutils{

/**
 * XmlWriter
 * base class for XML writer objects
 */
class XmlWriter : public XmlWrapper<xmlTextWriter>{
public:
    
    /**
     * start the document
     */
    void startDocument(){
        int rc = xmlTextWriterStartDocument(m_ptr,version(),encoding(),0);
        if (rc < 0){
            throw XmlConfigurationException("Error starting document.");
        }
    }
    /**
     * end the document
     */
    void endDocument(){
        int rc = xmlTextWriterEndDocument(m_ptr);
        if (rc < 0){
            throw XmlConfigurationException("Error ending document.");
        }
    }
    /**
     * start an element
     * @param name [IN] the element name
     */
    void startElement(const char* name){
        int rc = xmlTextWriterStartElement(m_ptr,xml_string(name));
        if (rc < 0){
            CONFIGURATION_EXCEPTION("Error starting element " << name);
        }
    }
    /**
     * end an element
     */
    void endElement(){
        int rc = xmlTextWriterEndElement(m_ptr);
        if (rc < 0){
            throw XmlConfigurationException("Error ending element.");
        }
    }
    /**
     * write an attribute
     * @param name [IN] attribute name
     * @param value [IN] attribute value
     */ 
    void writeAttribute(const char* name,const char* value){
        int rc = xmlTextWriterWriteAttribute(m_ptr,xml_string(name),xml_string(value));
        if (rc < 0){
            CONFIGURATION_EXCEPTION("Error writing attribute " << name);
        }
    }
    /**
     * write an element
     * @param name [IN] element name
     * @param value [IN] element value
     */ 
    void writeElement(const char* name,const char* value){
        int rc = xmlTextWriterWriteElement(m_ptr,xml_string(name),xml_string(value));
        if (rc < 0){
            CONFIGURATION_EXCEPTION("Error writing element " << name);
        }
    }
    /**
     * start a comment
     */ 
    void startComment(){
        int rc = xmlTextWriterStartComment(m_ptr);
        if (rc < 0){
            CONFIGURATION_EXCEPTION("Error starting comment");
        }
    }
    /**
     * end a comment
     */ 
    void endComment(){
        int rc = xmlTextWriterEndComment(m_ptr);
        if (rc < 0){
            CONFIGURATION_EXCEPTION("Error ending comment");
        }
    }
    
    /**
     * destructor
     */
    virtual ~XmlWriter(){}
        
protected:
    /**
     * Constructor
     * @param ptr [IN] the xmlTextWriter pointer
     */
    XmlWriter(xmlTextWriter * ptr): XmlWrapper<xmlTextWriter>(ptr,xmlFreeTextWriter){
    }
    /**
     * The encoding
     */
    const char* encoding(){
        return "UTF-8";
    }
    /**
     * The version
     */
    const char* version(){
        return "1.0";
    }
private:
    XmlWriter();
};    
    
/**
 * XmlFileWriter
 */
class XmlFileWriter : public XmlWriter{
public:
    /**
     * Constructor.
     * @param filename [IN] output file name.
     */
    XmlFileWriter(const std::string& filename) : XmlWriter(
        xmlNewTextWriterFilename(filename.c_str(),0)){
        if (0 == m_ptr){
            CONFIGURATION_EXCEPTION("Error creating XmlFileWriter for file '" << filename << "'");
        }
        // set indentation (don't check return codes)
        xmlTextWriterSetIndent(m_ptr,1);
        xmlTextWriterSetIndentString(m_ptr,xml_string("  "));
    }
    /**
     * destructor
     */
    virtual ~XmlFileWriter(){}
};

/**
 * Buffer for memory writers
 */
class Buffer : public XmlWrapper<xmlBuffer>{
public:
    /**
     * Constructor
     */
    Buffer() : XmlWrapper<xmlBuffer>(xmlBufferCreate(),xmlBufferFree){
        if (0 == m_ptr){
            throw XmlConfigurationException("Error creating buffer.");
        }
    }
    /**
     * Destructor
     */
    virtual ~Buffer(){}
protected:
};

class XmlMemoryWriter : public XmlWriter{
public:
    /**
     * Constructor
     * @param buffer [IN] the buffer to write to
     */
    XmlMemoryWriter(Buffer& buffer) : 
        XmlWriter(xmlNewTextWriterMemory(buffer.get(),0)){
        if (0 == m_ptr){
            throw XmlConfigurationException("Error creating XmlMemoryWriter.");
        }
    }
    /**
     * destructor
     */
    virtual ~XmlMemoryWriter(){}
};

} // namespace xmlutils
} // namespace config
} // namespace glite

#   else
#       error "libxml2 compiled without writer or output support"
#   endif

#endif // GLITE_DATA_CONFIG_XMLUTILS_WRITER_HXX_
