/**
 *  Copyright (c) Members of the EGEE Collaboration. 2004-2008.
 *  See http://www.eu-egee.org/partners/ for details on the copyright
 *  holders.
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 *  String Converter utility test case
 *
 *  Authors: Paolo Tedesco <paolo.tedesco@cern.ch>
 *  Version info: $Id: StringConverterTestCase.cxx,v 1.3 2009/12/03 17:18:01 molnarzs Exp $
 *  Release: $Name:  $
 *
 */

#include <cppunit/TestCase.h>
#include <cppunit/extensions/HelperMacros.h>
#include <glite/data/config/service/StringConverter.hxx>

#define CHECK_VALUE(value,expected) \
    if (value != expected){ \
        std::stringstream err; \
        err << #value << " : got '" << value << "', expected '" << expected << "'"; \
        CPPUNIT_FAIL(err.str().c_str()); \
    }

using namespace glite::config;
 
namespace{
    
struct FirstEnum{
    enum Value{
        zero,
        one,
        two
    };
};
struct SecondEnum{
    enum Value{
        a,
        b,
        c
    };
};

struct InvalidValueException{
};

}  // namespace

namespace glite{
namespace config{

template <>
struct StringConverter<FirstEnum::Value> : public EnumConverter<FirstEnum::Value>{
    StringConverter() : EnumConverter<FirstEnum::Value>(FirstEnum::zero,"zero"){
        add (FirstEnum::zero,   "zero");
        add (FirstEnum::one,    "one");
        add (FirstEnum::two,    "two");
    }
};
template <>
struct StringConverter<SecondEnum::Value> : public EnumConverter<SecondEnum::Value>{
    StringConverter(){
        add (SecondEnum::a, "a");
        add (SecondEnum::b, "b");
        add (SecondEnum::c, "c");
    }
protected:
    virtual std::string get_default_string(){
        throw InvalidValueException();
    }
    virtual SecondEnum::Value get_default_value(){
        throw InvalidValueException();
    }
};

} // config
} // glite

/**
 * StringConverterTestCase
 */
class StringConverterTestCase : public CppUnit::TestCase{
public:
protected:
    
    // declare tests
    CPPUNIT_TEST_SUITE( StringConverterTestCase );
        CPPUNIT_TEST( test_convert_int_01 );
        CPPUNIT_TEST( test_convert_int_02 );
        CPPUNIT_TEST( test_convert_bool_01 );
        CPPUNIT_TEST( test_convert_bool_02 );
        CPPUNIT_TEST( test_convert_bool_03 );
        CPPUNIT_TEST( test_convert_bool_04 );
        CPPUNIT_TEST( test_convert_bool_05 );
        CPPUNIT_TEST( test_convert_bool_06 );
        CPPUNIT_TEST( test_convert_enum_01 );
        CPPUNIT_TEST( test_convert_enum_02 );
        CPPUNIT_TEST( test_convert_enum_03 );
        CPPUNIT_TEST( test_convert_enum_04 );
        CPPUNIT_TEST( test_convert_enum_05 );
        CPPUNIT_TEST( test_convert_enum_06 );
        CPPUNIT_TEST( test_convert_string );
    CPPUNIT_TEST_SUITE_END();

    void test_convert_int_01(){
        CHECK_VALUE(StringConverter<int>::from_string("10"),10);
    }
    void test_convert_int_02(){
        CHECK_VALUE(StringConverter<int>::to_string(10),"10");
    }
    void test_convert_bool_01(){
        CHECK_VALUE(StringConverter<bool>::from_string("TRUE"),true);
        CHECK_VALUE(StringConverter<bool>::from_string("TrUe"),true);
        CHECK_VALUE(StringConverter<bool>::from_string("true"),true);
    }
    void test_convert_bool_02(){
        CHECK_VALUE(StringConverter<bool>::from_string("YES"),true);
        CHECK_VALUE(StringConverter<bool>::from_string("yEs"),true);
        CHECK_VALUE(StringConverter<bool>::from_string("yes"),true);
    }
    void test_convert_bool_03(){
        CHECK_VALUE(StringConverter<bool>::from_string("FALSE"),false);
        CHECK_VALUE(StringConverter<bool>::from_string("fALSe"),false);
        CHECK_VALUE(StringConverter<bool>::from_string("false"),false);
    }
    void test_convert_bool_04(){
        CHECK_VALUE(StringConverter<bool>::from_string("NO"),false);
        CHECK_VALUE(StringConverter<bool>::from_string("No"),false);
        CHECK_VALUE(StringConverter<bool>::from_string("no"),false);
    }
    void test_convert_bool_05(){
        CHECK_VALUE(StringConverter<bool>::to_string(true),"true");
    }
    void test_convert_bool_06(){
        CHECK_VALUE(StringConverter<bool>::to_string(false),"false");
    }
    void test_convert_enum_01(){
        CHECK_VALUE(StringConverter<FirstEnum::Value>::from_string("zero"),FirstEnum::zero);
        CHECK_VALUE(StringConverter<FirstEnum::Value>::from_string("one"),FirstEnum::one);
        CHECK_VALUE(StringConverter<FirstEnum::Value>::from_string("two"),FirstEnum::two);
    }
    void test_convert_enum_02(){
        CHECK_VALUE(StringConverter<FirstEnum::Value>::from_string("x"),FirstEnum::zero);
    }
    void test_convert_enum_03(){
        CHECK_VALUE(StringConverter<FirstEnum::Value>::to_string(FirstEnum::zero),"zero");
        CHECK_VALUE(StringConverter<FirstEnum::Value>::to_string(FirstEnum::one),"one");
        CHECK_VALUE(StringConverter<FirstEnum::Value>::to_string(FirstEnum::two),"two");
    }
    void test_convert_enum_04(){
        CHECK_VALUE(StringConverter<FirstEnum::Value>::to_string((FirstEnum::Value)-1),"zero");
    }
    void test_convert_enum_05(){
        try{
            StringConverter<SecondEnum::Value>::from_string("x");
            CPPUNIT_FAIL("Exception expected.");
        } catch (const InvalidValueException&){
            // ok
        }
    }
    void test_convert_enum_06(){
        try{
            StringConverter<SecondEnum::Value>::to_string((SecondEnum::Value)-1);
            CPPUNIT_FAIL("Exception expected.");
        } catch (const InvalidValueException&){
            // ok
        }
    }

    void test_convert_string()
    {
        std::string fixture("   a string with\t white \n spaces ");
        CHECK_VALUE(StringConverter<std::string>::to_string(fixture), fixture);
        CHECK_VALUE(StringConverter<std::string>::from_string(fixture), fixture);
    }
};

CPPUNIT_TEST_SUITE_NAMED_REGISTRATION( StringConverterTestCase, "StringConverterTestCase" );
