/**
 * Copyright (c) Members of the EGEE Collaboration. 2004.
 * See http://www.eu-egee.org/partners/ for details on the copyright holders.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *     http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 *      test-service.cxx
 *
 *      Authors: Paolo Badino <paolo.badino@cern.ch>, Paolo Tedesco <paolo.tedesco@cern.ch>
 *      Version info: $Id: test-service-hang.cxx,v 1.4 2008/12/17 13:39:15 ptedesco Exp $
 *      Release: $Name:  $
 */

#include "glite/data/config/service/ServiceConfiguration.hxx"
#include "glite/data/config/service/ParamValue.hxx"
#include "glite/data/config/service/ParamReader.hxx"

#include "test-module.hxx"
#include "log4cpp/Category.hh"

#include <signal.h>
#include <sys/wait.h>
#include <unistd.h>

using namespace log4cpp;
using namespace glite::config;

class TestServiceHang: public ServiceConfiguration {
public:
    /**
     * Initialize the Component
     */
    virtual int init(const Params& params);

    /**
     * Configure the Component
     */
    virtual int config(const Params& params);

    /**
     * Finailize the Component
     */
    virtual int fini();

    /**
     * Return Singleton Instance
     */
    static TestServiceHang * instance();
    
    /**
     * Finalize Singleton Instance
     */
    static void finalize();
private:
    /**
     * Constructor
     */
    TestServiceHang();
    
    /**
     * Destructor
     */
    virtual ~TestServiceHang(){}
    
    /**
     * Service Main Method
     */
    virtual int svc();

    /**
     * Callback to interrupt the Service Thread
     */
    virtual int svc_stop();
    
private:
    /**
     * Singleton Instance
     */
    static TestServiceHang * _instance;
    
    /**
     * The Logger
     */
    class Category& _logger;
    
    int          _interval;
    sig_atomic_t _run;
};

extern "C" {

/**
 * Create Component instance
 */
ComponentConfiguration * create_glite_component(){
    return TestServiceHang::instance();
}

/**
 * Destroy Component instance
 */
void destroy_glite_component(ComponentConfiguration * component){
    TestServiceHang::finalize();
}
} // End extern "C"

TestServiceHang* TestServiceHang::_instance = 0;

/**
 * Return Singleton Instance
 */
TestServiceHang * TestServiceHang::instance(){
    if(0 == _instance){
        _instance = new TestServiceHang();
    }
    return _instance;
}

/**
 * Finalize Singleton Instance
 */
void TestServiceHang::finalize(){
    if(0 != _instance){
        delete _instance;
    }
}

TestServiceHang::TestServiceHang():ServiceConfiguration("test-service-hang"),
   _logger(Category::getInstance("test-service-hang")),_interval(10),_run(0){
}

/**
 * Initialize the Component
 */
int TestServiceHang::init(const Params& params){
    
    Params::const_iterator it;
    for(it = params.begin(); it != params.end(); it++){
        ParamValue * val = dynamic_cast<ParamValue *>((*it).second);
        _logger.debug("Init Parameter %s value is %s",(*it).first.c_str(),val->getValue().c_str());
    }
    _logger.info("initialization completed");
    
    return 0;
}

/**
 * Configure the Component
 */
int TestServiceHang::config(const Params& params){
    ParamReader reader(getName(),params);
    reader.getParameter("interval",_interval)
        .required()
        .gt(0);
    _logger.info("configuration completed");
    
    return 0;
}

/**
 * Finailize the Component
 */
int TestServiceHang::fini(){
    
    _logger.info("finalization completed");
    
    return 0;
}

void glite_doSomethingElse(const char * requestor, const char * str);

/**
 * TestServiceHang Main
 */
int TestServiceHang::svc(){
    
    _run = 1;
    _logger.info("Service Started");
    
    int count = 0;
    while(1){ // Always run
    
        _logger.info("Iteration #%d", count);
       
        TestComponent::instance()->doSomething("test-service","something");
        glite_doSomethingElse("test-service","hello world");
        sleep(_interval);
        count++;
    }
    _logger.info("Service Stopped");
    
    return 0;
}

/**
 * Stop TestServiceHang Main
 */
int TestServiceHang::svc_stop(){
    
    _logger.info("Stopping Service");
    _run = 0;
    
    return 0;
}

