/*
 * Copyright (c) Members of the EGEE Collaboration. 2004.
 * See http://www.eu-egee.org/partners/ for details on the copyright holders.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *     http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 *      test-service.cxx
 *
 *      Authors: Paolo Badino <paolo.badino@cern.ch>
 *      Version info: $Id: test-service.cxx,v 1.4 2009/10/29 10:22:44 szamsu Exp $
 *      Release: $Name:  $
 *
 *      Revision history:
 *      $Log: test-service.cxx,v $
 *      Revision 1.4  2009/10/29 10:22:44  szamsu
 *      Amending the last commit for bug#56533 by updates from Adrian Stelmaszyk <adisan@man.poznan.pl>
 *
 *      Revision 1.3  2007/12/19 11:35:24  ptedesco
 *      logs and build files cleanup
 *
 *      Revision 1.2  2004/08/30 12:48:10  badino
 *      Add ParamValue Class
 *      Set Version to 0.1.0
 *
 *      Revision 1.1.1.1  2004/08/18 15:52:41  badino
 *      Import of Module GLite Data config Service v.0.0
 *
 *
 */

#include "glite/data/config/service/ServiceConfiguration.hxx"
#include "glite/data/config/service/ParamValue.hxx"

#include "test-module.hxx"
#include "log4cpp/Category.hh"

#include <signal.h>
#include <sys/wait.h>
#include <unistd.h>
#include <cstring>
#include <cstdlib>

using namespace log4cpp;
using namespace glite::config;

class TestService: public ServiceConfiguration {
public:
    /**
     * Initialize the Component
     */
    virtual int init(const Params& params);

    /**
     * Configure the Component
     */
    virtual int config(const Params& params);

    /**
     * Finailize the Component
     */
    virtual int fini();

    /**
     * Return Singleton Instance
     */
    static TestService * instance();
    
    /**
     * Finalize Singleton Instance
     */
    static void finalize();
private:
    /**
     * Constructor
     */
    TestService();
    
    /**
     * Destructor
     */
    virtual ~TestService(){}
    
    /**
     * Service Main Method
     */
    virtual int svc();

    /**
     * Callback to interrupt the Service Thread
     */
    virtual int svc_stop();
    
private:
    /**
     * Singleton Instance
     */
    static TestService * _instance;
    
    /**
     * The Logger
     */
    class Category& _logger;
    
    int          _interval;
    sig_atomic_t _run;
};

extern "C" {

/**
 * Create Component instance
 */
ComponentConfiguration * create_glite_component(){
    return TestService::instance();
}

/**
 * Destroy Component instance
 */
void destroy_glite_component(ComponentConfiguration * component){
    TestService::finalize();
}
} // End extern "C"

TestService* TestService::_instance = 0;

/**
 * Return Singleton Instance
 */
TestService * TestService::instance(){
    if(0 == _instance){
        _instance = new TestService();
    }
    return _instance;
}

/**
 * Finalize Singleton Instance
 */
void TestService::finalize(){
    if(0 != _instance){
        delete _instance;
    }
}

TestService::TestService():ServiceConfiguration("test-service"),
   _logger(Category::getInstance("test-service")),_interval(10),_run(0){
}

/**
 * Initialize the Component
 */
int TestService::init(const Params& params){
    
    Params::const_iterator it;
    for(it = params.begin(); it != params.end(); it++){
        ParamValue * val = dynamic_cast<ParamValue *>((*it).second);
        _logger.debug("Init Parameter %s value is %s",(*it).first.c_str(),val->getValue().c_str());
    }
    _logger.info("initialization completed");
    
    return 0;
}

/**
 * Configure the Component
 */
int TestService::config(const Params& params){
    
    Params::const_iterator it;
    for(it = params.begin(); it != params.end(); it++){
        ParamValue * val = dynamic_cast<ParamValue *>((*it).second);
        _logger.debug("Config Parameter %s value is %s",(*it).first.c_str(),val->getValue().c_str());
        if(0 == strcmp((*it).first.c_str(),"interval")){
            _interval = atoi(val->getValue().c_str());
            if(_interval <= 0){
                _interval = 1;
            }
        }
    }
    _logger.info("configuration completed");
    
    return 0;
}

/**
 * Finailize the Component
 */
int TestService::fini(){
    
    _logger.info("finalization completed");
    
    return 0;
}

void glite_doSomethingElse(const char * requestor, const char * str);

/**
 * TestService Main
 */
int TestService::svc(){
    
    _run = 1;
    _logger.info("Service Started");
    
    int count = 0;
    while(_run){
    
        _logger.info("Iteration #%d", count);
       
        TestComponent::instance()->doSomething("test-service","something");
        glite_doSomethingElse("test-service","hello world");
        sleep(_interval);
        count++;
    }
    _logger.info("Service Stopped");
    
    return 0;
}

/**
 * Stop TestService Main
 */
int TestService::svc_stop(){
    
    _logger.info("Stopping Service");
    _run = 0;
    
    return 0;
}

