/*
 * Copyright (c) Members of the EGEE Collaboration. 2004.
 * See http://www.eu-egee.org/partners/ for details on the copyright holders.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *     http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * test-static.cxx
 *
 * Authors: Paolo Badino <paolo.badino@cern.ch>
 * Version info: $Id: test-static.cxx,v 1.4 2009/10/29 10:22:44 szamsu Exp $
 * Release: $Name:  $
 *
 */

#include "glite/data/config/service/ComponentConfiguration.hxx"
#include "glite/data/config/service/ParamValue.hxx"

#include "log4cpp/Category.hh"

#include <signal.h>
#include <sys/wait.h>
#include <unistd.h>
#include "glite/data/config/service/glite-config.h"
#include "glite/data/config/service/glite-config-register.hxx"
#include <cstring>

using namespace log4cpp;
using namespace glite::config;

class TestStaticComponent: public ComponentConfiguration {
public:
    /**
     * Initialize the Component
     */
    virtual int init(const Params& params);

    /**
     * Configure the Component
     */
    virtual int config(const Params& params);

    /**
     * Start the Component
     */
    virtual int start();

    /**
     * Stop the Component
     */
    virtual int stop();
 
    /**
     * Finailize the Component
     */
    virtual int fini();

    /**
     * Return Singleton Instance
     */
    static TestStaticComponent * instance();
    
    /**
     * Finalize Singleton Instance
     */
    static void finalize();
    
    /**
     * Return Config Parameter Value
     */
    const char * getParam() const { return _param.c_str();}
    
private:
    /**
     * Constructor
     */
    TestStaticComponent();
    
    /**
     * Destructor
     */
    virtual ~TestStaticComponent(){}
    
private:
    /**
     * Singleton Instance
     */
    static TestStaticComponent * _instance;
    
    /**
     * The Logger
     */
    class Category& _logger;
        
    /**
     * The Config Parameter
     */
    std::string _param;
};

TestStaticComponent* TestStaticComponent::_instance = 0;

/**
 * Return Singleton Instance
 */
TestStaticComponent * TestStaticComponent::instance(){
    if(0 == _instance){
        _instance = new TestStaticComponent();
    }
    return _instance;
}

/**
 * Finalize Singleton Instance
 */
void TestStaticComponent::finalize(){
    if(0 != _instance){
        delete _instance;
    }
}

TestStaticComponent::TestStaticComponent():ComponentConfiguration("test-static"),
   _logger(Category::getInstance("test-static")){
}

/**
 * Initialize the Component
 */
int TestStaticComponent::init(const Params& params){
    
    Params::const_iterator it;
    for(it = params.begin(); it != params.end(); it++){
        ParamValue * val = dynamic_cast<ParamValue *>((*it).second);
        if(0 != val){
            _logger.debug("Init Parameter %s value is %s",val->getName(),val->getValue().c_str());
            
            if(0 == strcasecmp(val->getName(),"param")){
                _param = val->getValue();
            }
        }
    }
    _logger.info("initialization completed");
    
    return 0;
}

/**
 * Configure the Component
 */
int TestStaticComponent::config(const Params& params){
    
    Params::const_iterator it;
    for(it = params.begin(); it != params.end(); it++){
        ParamValue * val = dynamic_cast<ParamValue *>((*it).second);
        if(0 != val){
            _logger.debug("Config Parameter %s value is %s",val->getName(),val->getValue().c_str());
        }
    }
    _logger.info("configuration completed");
    
    return 0;
}

/**
 * Finailize the Component
 */
int TestStaticComponent::fini(){
    
    _logger.info("finalization completed");
    
    return 0;
}

/**
 * TestStaticComponent start
 */
int TestStaticComponent::start(){
    
    return 0;
}

/**
 * Stop TestStaticComponent
 */
int TestStaticComponent::stop(){
    
    return 0;
}

int main(int argc, char * argv[]){
    const char * config_name = "test-config-service-static";
    
    // Register the Static Component
    glite_config_register(TestStaticComponent::instance());
    fprintf(stdout,"Component %s Registered into Service %s\n",TestStaticComponent::instance()->getName(), config_name);
        
    // Initialize the Configuration
    int initres = glite_config_initialize(config_name);
    if(initres < 0){
        fprintf(stderr,"Cannot Initialize with Config Name %s!\n",config_name);
        return -1;
    }  
    fprintf(stdout,"Service %s initialized\n",config_name);
    
    // get a Parameter value
    const char * param = TestStaticComponent::instance()->getParam();
    fprintf(stdout,"Got Param %s\n",param);
    
    // Finalize
    glite_config_finalize();
    fprintf(stdout,"Service %s Finalized\n",config_name);
    
    return 0;
}
